# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Union

from mmengine.device import get_device
from mmengine.logging import MMLogger
from mmengine.model import BaseModel
from mmengine.runner import Runner


class DeployTestRunner(Runner):
    """The runner for test models.

    Args:
        log_file (str | None): The path of log file. Default is ``None``.
        device (str): The device type.
    """

    def __init__(self,
                 log_file: Optional[str] = None,
                 device: str = get_device(),
                 *args,
                 **kwargs):

        self._log_file = log_file
        self._device = device
        super(DeployTestRunner, self).__init__(*args, **kwargs)

    def wrap_model(self, model_wrapper_cfg: Optional[Dict],
                   model: BaseModel) -> BaseModel:
        """Wrap the model to :obj:``MMDistributedDataParallel`` or other custom
        distributed data-parallel module wrappers.

        An example of ``model_wrapper_cfg``::

            model_wrapper_cfg = dict(
                broadcast_buffers=False,
                find_unused_parameters=False
            )

        Args:
            model_wrapper_cfg (dict, optional): Config to wrap model. If not
                specified, ``DistributedDataParallel`` will be used in
                distributed environment. Defaults to None.
            model (BaseModel): Model to be wrapped.

        Returns:
            BaseModel or DistributedDataParallel: BaseModel or subclass of
            ``DistributedDataParallel``.
        """
        return model.to(self._device)

    def build_logger(self,
                     log_level: Union[int, str] = 'INFO',
                     log_file: str = None,
                     **kwargs) -> MMLogger:
        """Build a global accessible MMLogger.

        Args:
            log_level (int or str): The log level of MMLogger handlers.
                Defaults to 'INFO'.
            log_file (str, optional): Path of filename to save log.
                Defaults to None.
            **kwargs: Remaining parameters passed to ``MMLogger``.

        Returns:
            MMLogger: A MMLogger object build from ``logger``.
        """
        if log_file is None:
            log_file = self._log_file

        return super().build_logger(log_level, log_file, **kwargs)



# Copyright (c) OpenMMLab. All rights reserved.
import functools
import inspect
import types
import warnings
from abc import ABCMeta, abstractmethod
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import juxtapose.mmdeploy
from juxtapose.mmdeploy.utils.constants import IR, Backend


def eval_with_import(path: str) -> Any:
    """Evaluate the string as Python script.

    Args:
        path (str): The path to evaluate.

    Returns:
        Any: The result of evaluation.
    """
    split_path = path.split(".")
    for i in range(len(split_path), 0, -1):
        try:
            exec("import {}".format(".".join(split_path[:i])))
            break
        except Exception:
            continue
    return eval(path)


def import_function(path: str) -> Tuple[Callable, Optional[type]]:
    """Import and evaluate a function. If the function is defined in a class,
    evaluate the class additionally.

    Args:
        path (str): The path to evaluate.

    Returns:
        Callable: The function of evaluation.
        type: The class of evaluation if the function is defined in a class, or
            None.
    """
    split_path = path.split(".")
    for i in range(len(split_path), 0, -1):
        try:
            exec("import {}".format(".".join(split_path[:i])))
            break
        except Exception:
            continue

    obj = eval(path)

    # The path that might be a class
    previous_obj = eval(".".join(split_path[:-1]))

    # Check if the path leads to a class
    if inspect.isclass(previous_obj):
        return obj, previous_obj
    else:
        return obj, None


def collect_env(backend: Backend, ir: IR, **kwargs) -> Dict:
    """Collect current environment information, including backend, ir, codebase
    version, etc. Rewriters will be checked according to env infos.

    Args:
        backend (Backend): Current backend.
        ir (IR): Current IR.

    Returns:
        Dict: Record the value of Backend and IR as well as the versions of
        libraries.
    """
    from juxtapose.mmdeploy.utils import get_backend_version, get_codebase_version

    env = dict(backend=backend, ir=ir)
    env["mmdeploy"] = juxtapose.mmdeploy.__version__
    env.update(get_backend_version())
    env.update(get_codebase_version())
    env.update(kwargs)
    return env


class Checker(metaclass=ABCMeta):
    """The interface for checking whether a rewriter is valid."""

    def __init__(self):
        pass

    @abstractmethod
    def check(self, env: Dict) -> bool:
        """Check the if the rewriter is valid according to environment.

        Args:
            env (Dict): The backend, IR info and version info.
        """
        pass


class BackendChecker(Checker):
    """Checker that determines which backend the rewriter must run on.

    Args:
        required_backend (Backend): The rewriter will be activated on
            which backend.
    """

    def __init__(self, required_backend: Backend):
        super().__init__()
        self.required_backend = required_backend

    def check(self, env: Dict) -> bool:
        """Check the if the rewriter is valid according to backend.

        Args:
            env (Dict): The backend, IR info and version info.
        """
        return env["backend"] == self.required_backend


class IRChecker(Checker):
    """Checker that determines which IR the rewriter must run on.

    Args:
        required_ir (IR): The rewriter will be activated on which IR.
    """

    def __init__(self, required_ir: IR):
        super().__init__()
        self.required_ir = required_ir

    def check(self, env: Dict) -> bool:
        """Check the if the rewriter is valid according to IR.

        Args:
            env (Dict): The backend, IR info and version info.
        """
        return env["ir"] == self.required_ir


class LibVersionChecker(Checker):
    """Checker that determines which IR the rewriter must run on.

    Args:
        lib (str): The name of library.
        min_version (str | None): The rewriter should no lower than which
            version. Default to `None`.
        max_version (str | None): The rewriter should no greater than which
            version. Default to `None`.
    """

    def __init__(
        self,
        lib: str,
        min_version: Optional[str] = None,
        max_version: Optional[str] = None,
    ):
        super().__init__()
        self.lib = lib
        self.min_version = min_version
        self.max_version = max_version

    def check(self, env: Dict) -> bool:
        """Check the if the rewriter is valid according to library version.

        Args:
            env (Dict): The backend, IR info and version info.
        """
        # If the library has not been installed
        if env[self.lib] is None:
            return False

        from packaging import version

        valid = True
        # The version should no less than min version and no greater than
        # max version.
        if self.min_version is not None:
            if version.parse(env[self.lib]) < version.parse(self.min_version):
                valid = False
        if self.max_version is not None:
            if version.parse(env[self.lib]) > version.parse(self.max_version):
                valid = False
        return valid


class RewriterRegistry:
    """A registry that records rewrite objects.

    Logically this class is a two-dimensional table which maintains an object
    list for each backend. The records can be inserted to this table through
    register.

    Members:
        _rewrite_records (Dict[Backend, Dict[str, Dict]]): A data structure
            which records the register message in a specific backend.

    Example:
        >>> FUNCTION_REGISTRY = RewriterRegistry()
        >>> @FUNCTION_REGISTRY.register_object(backend="default")
        >>> def add():
        >>>     return a + b
        >>> records = FUNCTION_REGISTRY.get_record("default")
    """

    def __init__(self):
        self._rewrite_records = dict()

    def get_records(self, env: Dict) -> List:
        """Get all registered records that are valid in the given environment
        from record table.

        If the backend and IR of rewriter are set to 'default', then the
        rewriter is regarded as default rewriter. The default rewriter will be
        activated only when all other rewriters are not valid. If there are
        multiple rewriters are valid (except default rewriter), we will
        activate the first one (The order is determined by the time when
        rewriters are loaded).

        Args:
            env (dict): Environment dictionary that includes backend, IR,
                codebase version, etc.

        Returns:
            List: A list that includes valid records.
        """
        default_records = list()
        records = list()

        for origin_function, rewriter_records in self._rewrite_records.items():
            default_rewriter = None
            final_rewriter = None
            for record in rewriter_records:
                # Get the checkers of current rewriter
                checkers: List[Checker] = record["_checkers"]

                # Check if the rewriter is default rewriter
                if len(checkers) == 0:
                    #  Process the default rewriter exceptionally
                    if default_rewriter is None:
                        default_rewriter = record
                    else:
                        warnings.warn(
                            "Detect multiple valid rewriters for "
                            f"{origin_function}, use the first rewriter."
                        )
                else:
                    # Check if the checker is valid.
                    # The checker is valid only if all the checks are passed
                    valid = True
                    for checker in checkers:
                        if not checker.check(env):
                            valid = False
                            break

                    if valid:
                        # Check if there are multiple valid rewriters
                        if final_rewriter is not None:
                            warnings.warn(
                                "Detect multiple valid rewriters for"
                                f"{origin_function}, use the first rewriter."
                            )
                        else:
                            final_rewriter = record

            # Append final rewriter.
            # If there is no valid rewriter, try not apply default rewriter
            if final_rewriter is not None:
                records.append((origin_function, final_rewriter))
            elif default_rewriter is not None:
                default_records.append((origin_function, default_rewriter))

        # Make the default records como to the front of list because we may
        # want the non-default records to override them.
        return default_records + records

    def _register(
        self,
        name: str,
        backend: Backend,
        ir: IR,
        extra_checkers: List[Checker],
        **kwargs,
    ):
        """The implementation of register."""

        # Merge checkers to kwargs
        record_dict = kwargs

        # Try to create a checker according to 'backend' field
        if backend != Backend.DEFAULT:
            extra_checkers.append(BackendChecker(backend))

        # Try to create a checker according to 'ir' field
        if ir != IR.DEFAULT:
            extra_checkers.append(IRChecker(ir))

        record_dict["_checkers"] = extra_checkers

        # There may be multiple rewriters of a function/module. We use a list
        # to store the rewriters of a function/module.
        if name not in self._rewrite_records:
            self._rewrite_records[name] = list()
        self._rewrite_records[name].append(record_dict)

    def register_object(
        self,
        name: str,
        backend: str,
        ir: IR,
        extra_checkers: Optional[Union[Checker, List[Checker]]] = None,
        **kwargs,
    ) -> Callable:
        """The decorator to register an object.

        Args:
            name (str): The import path to access the function/module.
            backend (str): The rewriter will be activated on which backend.
            ir (IR): The rewriter will be activated on which ir.
            extra_checkers (None | Checker | List[Checker]): Other requirements
                for the rewriters. Default to `None`.

        Returns:
            Callable: The decorator.
        """

        if extra_checkers is None:
            extra_checkers = []
        elif isinstance(extra_checkers, Checker):
            extra_checkers = [extra_checkers]

        backend = Backend.get(backend)

        def decorator(object):
            self._register(name, backend, ir, extra_checkers, _object=object, **kwargs)
            return object

        return decorator

    def remove_record(self, object: Any, filter_cb: Optional[Callable] = None):
        """Remove record.

        Args:
            object (Any): The object to remove.
            filter_cb (Callable): Check if the object need to be remove.
                Defaults to None.
        """
        key_to_pop = []
        for key, records in self._rewrite_records.items():
            for rec in records:
                if rec["_object"] == object:
                    if filter_cb is not None:
                        if filter_cb(rec):
                            continue
                    key_to_pop.append((key, rec))

        for key, rec in key_to_pop:
            records = self._rewrite_records[key]
            records.remove(rec)
            if len(records) == 0:
                self._rewrite_records.pop(key)


class ContextCaller:
    """A callable object used in RewriteContext.

    This class saves context variables as member variables. When a rewritten
    function is called in RewriteContext, an instance of this class will be
    passed as the first argument of the function.

    Args:
        func (Callable): The rewritten function to call.
        origin_func (Callable): The function that is going to be rewritten.
            Note that in symbolic function origin_func may be 'None'.
        cfg (Dict): The deploy config dictionary.

    Example:
        >>> @FUNCTION_REWRITER.register_rewriter(func_name='torch.add')
        >>> def func(x, y):
        >>>     # ctx is an instance of ContextCaller
        >>>     ctx = FUNCTION_REWRITER.get_context()
        >>>     print(ctx.cfg)
        >>>     return x + y
    """

    def __init__(self, func: Callable, origin_func: Callable, cfg: Dict, **kwargs):
        self.func = func
        self.origin_func = origin_func
        self.cfg = cfg
        # PyTorch will do annotation check on symbolic function
        # Update the annotation so ContextCaller can pass the check.
        if origin_func is not None:
            wraps(origin_func)(self)
        else:
            self.__annotations__ = getattr(func, "__annotations__", {})

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __call__(self, *args, **kwargs):
        """Directly call self.func."""
        return self.func(self, *args, **kwargs)

    def get_wrapped_caller(self):
        """Generate a wrapped caller for function rewrite."""

        # Rewrite function should not call a member function, so we use a
        # wrapper to generate a Callable object.
        def wrapper(*args, **kwargs):
            # Add a new argument (context message) to function
            # Because "self.func" is a function but not a member function,
            # we should pass self as the first argument
            return self.func(self, *args, **kwargs)

        return wrapper


def get_func_qualname(func: Callable) -> str:
    """get function name."""
    assert isinstance(func, Callable), f"{func} is not a Callable object."
    _func_name = None
    if hasattr(func, "__qualname__"):
        _func_name = f"{func.__module__}.{func.__qualname__}"
    elif hasattr(func, "__class__"):
        _func_name = func.__class__
    else:
        _func_name = str(func)
    return _func_name


def get_frame_func(top: int = 1) -> Callable:
    """get func of frame."""
    frameinfo = inspect.stack()[top]
    frame = frameinfo.frame

    g_vars = frame.f_globals
    func_name = frameinfo.function
    assert func_name in g_vars, f"Can not find function: {func_name} in global."
    func = g_vars[func_name]
    return func


def get_frame_qualname(top: int = 1) -> str:
    """get frame name."""
    frameinfo = inspect.stack()[top]
    frame = frameinfo.frame

    g_vars = frame.f_globals
    func_name = frameinfo.function
    assert func_name in g_vars, f"Can not find function: {func_name} in global."
    func = g_vars[func_name]
    module_name = inspect.getmodule(func).__name__

    return f"{module_name}.{func_name}"


def copy_function(f: types.FunctionType):
    """Copy the function."""
    # copy the global so we can get different func for different origin
    glb = f.__globals__.copy()
    name = f.__name__
    g = types.FunctionType(
        f.__code__, glb, name=name, argdefs=f.__defaults__, closure=f.__closure__
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    glb[name] = g
    return g



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from functools import partial
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import torch

from juxtapose.mmdeploy.apis.core import PIPELINE_MANAGER
from juxtapose.mmdeploy.core import RewriterContext, patch_model
from juxtapose.mmdeploy.utils import IR, Backend, get_ir_config, get_root_logger
from .optimizer import *  # noqa
from .passes import optimize_onnx


@PIPELINE_MANAGER.register_pipeline()
def export(
    model: torch.nn.Module,
    args: Union[torch.Tensor, Tuple, Dict],
    output_path_prefix: str,
    backend: Union[Backend, str] = "default",
    input_metas: Optional[Dict] = None,
    context_info: Dict = dict(),
    input_names: Optional[Sequence[str]] = None,
    output_names: Optional[Sequence[str]] = None,
    opset_version: int = 11,
    dynamic_axes: Optional[Dict] = None,
    verbose: bool = False,
    keep_initializers_as_inputs: Optional[bool] = None,
    optimize: bool = False,
):
    """Export a PyTorch model into ONNX format. This is a wrap of
    `torch.onnx.export` with some enhancement.

    Examples:
        >>> from juxtapose.mmdeploy.apis.onnx import export
        >>>
        >>> model = create_model()
        >>> args = get_input_tensor()
        >>>
        >>> export(
        >>>     model,
        >>>     args,
        >>>     'place/to/save/model',
        >>>     backend='tensorrt',
        >>>     input_names=['input'],
        >>>     output_names=['output'],
        >>>     dynamic_axes={'input': {
        >>>         0: 'batch',
        >>>         2: 'height',
        >>>         3: 'width'
        >>>     }})

    Args:
        model (torch.nn.Module): the model to be exported.
        args (torch.Tensor|Tuple|Dict): Dummy input of the model.
        output_path_prefix (str): The output file prefix. The model will
            be saved to `<output_path_prefix>.onnx`.
        backend (Backend|str): Which backend will the graph be used. Different
            backend would generate different graph.
        input_metas (Dict): The constant inputs of the model.
        context_info (Dict): The information that would be used in the context
            of exporting.
        input_names (Sequence[str]): The input names of the model.
        output_names (Sequence[str]): The output names of the model.
        opset_version (int): The version of ONNX opset version. 11 as default.
        dynamic_axes (Dict): The information used to determine which axes are
            dynamic.
        verbose (bool): Enable verbose model on `torch.onnx.export`.
        keep_initializers_as_inputs (bool): Whether we should add inputs for
            each initializer.
        optimize (bool): Perform optimize on model.
    """
    output_path = output_path_prefix + ".onnx"

    logger = get_root_logger()
    logger.info(f"Export PyTorch model to ONNX: {output_path}.")

    def _add_or_update(cfg: dict, key: str, val: Any):
        if key in cfg and isinstance(cfg[key], dict) and isinstance(val, dict):
            cfg[key].update(val)
        else:
            cfg[key] = val

    context_info = deepcopy(context_info)
    deploy_cfg = context_info.pop("deploy_cfg", dict())
    ir_config = dict(
        type="onnx",
        input_names=input_names,
        output_names=output_names,
        opset_version=opset_version,
        dynamic_axes=dynamic_axes,
        verbose=verbose,
        keep_initializers_as_inputs=keep_initializers_as_inputs,
    )
    _add_or_update(deploy_cfg, "ir_config", ir_config)
    ir = IR.get(get_ir_config(deploy_cfg)["type"])
    if isinstance(backend, Backend):
        backend = backend.value
    backend_config = dict(type=backend)
    _add_or_update(deploy_cfg, "backend_config", backend_config)

    context_info["cfg"] = deploy_cfg
    context_info["ir"] = ir
    if "backend" not in context_info:
        context_info["backend"] = backend
    if "opset" not in context_info:
        context_info["opset"] = opset_version

    # patch model
    patched_model = patch_model(model, cfg=deploy_cfg, backend=backend, ir=ir)

    if "onnx_custom_passes" not in context_info:
        onnx_custom_passes = optimize_onnx if optimize else None
        context_info["onnx_custom_passes"] = onnx_custom_passes
    with RewriterContext(**context_info), torch.no_grad():
        # patch input_metas
        if input_metas is not None:
            assert isinstance(
                input_metas, dict
            ), f"Expect input_metas type is dict, get {type(input_metas)}."
            model_forward = patched_model.forward

            def wrap_forward(forward):
                def wrapper(*arg, **kwargs):
                    return forward(*arg, **kwargs)

                return wrapper

            patched_model.forward = wrap_forward(patched_model.forward)
            patched_model.forward = partial(patched_model.forward, **input_metas)

        torch.onnx.export(
            patched_model,
            args,
            output_path,
            export_params=True,
            input_names=input_names,
            output_names=output_names,
            opset_version=opset_version,
            dynamic_axes=dynamic_axes,
            keep_initializers_as_inputs=keep_initializers_as_inputs,
            verbose=verbose,
        )

        if input_metas is not None:
            patched_model.forward = model_forward



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Union

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.nn.functional.group_norm", backend="ncnn"
)
def group_norm__ncnn(
    input: torch.Tensor,
    num_groups: int,
    weight: Union[torch.Tensor, torch.NoneType] = None,
    bias: Union[torch.Tensor, torch.NoneType] = None,
    eps: float = 1e-05,
) -> torch.Tensor:
    """Rewrite `group_norm` for ncnn backend.

    InstanceNorm in ncnn require input with shape [C, H, W]. So we have to
    reshape the input tensor before it.
    """
    input_shape = input.shape
    batch_size = input_shape[0]
    # We cannot use input.reshape(batch_size, num_groups, -1, 1)
    # instead, or we will meet bug on ncnn Reshape ops.
    input_reshaped = input.reshape(batch_size, num_groups, -1)
    input_reshaped = input_reshaped.unsqueeze(3)
    # the weight_'s size is not the same as weight's size
    # we only use groupnorm to calculate instancenorm, but the
    # input parameters may not be the same, and need to transform.
    weight_ = torch.tensor([1.0] * num_groups).type_as(input)
    bias_ = torch.tensor([0.0] * num_groups).type_as(input)

    norm_reshaped = torch.nn.functional.instance_norm(
        input_reshaped, weight=weight_, bias=bias_, eps=eps
    )

    norm = norm_reshaped.reshape(*input_shape)
    if weight is None:
        weight = torch.tensor([1.0]).type_as(input)
    if bias is None:
        bias = torch.tensor([0.0]).type_as(input)
    weight = weight.reshape(1, -1, 1, 1)
    bias = bias.reshape(1, -1, 1, 1)

    return norm * weight + bias



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Optional, Sequence, Tuple, Union

import mmengine
import torch
from torch import Tensor

from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.core.rewriters.rewriter_utils import LibVersionChecker
from juxtapose.mmdeploy.utils import Backend, load_config


def get_post_processing_params(deploy_cfg: Union[str, mmengine.Config]):
    """Get mmdet post-processing parameters from config.

    Args:
        deploy_cfg (str | mmengine.Config): The path or content of config.

    Returns:
        dict: A dict of parameters for mmdet.
    """
    deploy_cfg = load_config(deploy_cfg)[0]
    codebase_key = "codebase_config"
    assert codebase_key in deploy_cfg
    codebase_config = deploy_cfg[codebase_key]
    post_params = codebase_config.get("post_processing", None)
    assert post_params is not None, "Failed to get `post_processing`."
    return post_params


def clip_bboxes(
    x1: Tensor,
    y1: Tensor,
    x2: Tensor,
    y2: Tensor,
    max_shape: Union[Tensor, Sequence[int]],
):
    """Clip bboxes for onnx.

    Since torch.clamp cannot have dynamic `min` and `max`, we scale the
      boxes by 1/max_shape and clamp in the range [0, 1] if necessary.

    Args:
        x1 (Tensor): The x1 for bounding boxes.
        y1 (Tensor): The y1 for bounding boxes.
        x2 (Tensor): The x2 for bounding boxes.
        y2 (Tensor): The y2 for bounding boxes.
        max_shape (Tensor | Sequence[int]): The (H,W) of original image.

    Returns:
        tuple(Tensor): The clipped x1, y1, x2, y2.
    """
    assert len(max_shape) == 2, "`max_shape` should be [h, w]"
    if isinstance(max_shape, torch.Tensor):
        # scale by 1/max_shape
        x1 = x1 / max_shape[1]
        y1 = y1 / max_shape[0]
        x2 = x2 / max_shape[1]
        y2 = y2 / max_shape[0]

        # clamp [0, 1]
        x1 = torch.clamp(x1, 0, 1)
        y1 = torch.clamp(y1, 0, 1)
        x2 = torch.clamp(x2, 0, 1)
        y2 = torch.clamp(y2, 0, 1)

        # scale back
        x1 = x1 * max_shape[1]
        y1 = y1 * max_shape[0]
        x2 = x2 * max_shape[1]
        y2 = y2 * max_shape[0]
    else:
        x1 = torch.clamp(x1, 0, max_shape[1])
        y1 = torch.clamp(y1, 0, max_shape[0])
        x2 = torch.clamp(x2, 0, max_shape[1])
        y2 = torch.clamp(y2, 0, max_shape[0])
    return x1, y1, x2, y2


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.codebase.mmdet.deploy.utils.clip_bboxes",
    backend="tensorrt",
    extra_checkers=LibVersionChecker("tensorrt", min_version="8.2"),
)
def clip_bboxes__trt8(
    x1: Tensor,
    y1: Tensor,
    x2: Tensor,
    y2: Tensor,
    max_shape: Union[Tensor, Sequence[int]],
):
    """Clip bboxes for onnx. From TensorRT 8 we can do the operators on the
    tensors directly.

    Args:
        ctx (ContextCaller): The context with additional information.
        x1 (Tensor): The x1 for bounding boxes.
        y1 (Tensor): The y1 for bounding boxes.
        x2 (Tensor): The x2 for bounding boxes.
        y2 (Tensor): The y2 for bounding boxes.
        max_shape (Tensor | Sequence[int]): The (H,W) of original image.

    Returns:
        tuple(Tensor): The clipped x1, y1, x2, y2.
    """
    assert len(max_shape) == 2, "`max_shape` should be [h, w]"
    x1 = torch.clamp(x1, 0, max_shape[1])
    y1 = torch.clamp(y1, 0, max_shape[0])
    x2 = torch.clamp(x2, 0, max_shape[1])
    y2 = torch.clamp(y2, 0, max_shape[0])
    return x1, y1, x2, y2


def pad_with_value(
    x: Tensor, pad_dim: int, pad_size: int, pad_value: Optional[Any] = None
):
    """Pad a tensor with a value along some dim.

    Args:
        x (Tensor): Input tensor.
        pad_dim (int): Along which dim to pad.
        pad_size (int): To which size to pad.
        pad_value (Any): Filled value for padding. Defaults to `None`.

    Returns:
        Tensor: Padded tensor.
    """
    x_shape = list(x.shape)
    pad_shape = x_shape[:pad_dim] + [pad_size] + x_shape[pad_dim + 1 :]
    x_pad = x.new_zeros(pad_shape)
    if pad_value is not None:
        x_pad = x_pad + pad_value
    x = torch.cat([x, x_pad], dim=pad_dim)
    return x


def pad_with_value_if_necessary(
    x: Tensor, pad_dim: int, pad_size: int, pad_value: Optional[Any] = None
) -> Tensor:
    """Pad a tensor with a value along some dim if necessary.

    Args:
        x (Tensor): Input tensor.
        pad_dim (int): Along which dim to pad.
        pad_size (int): To which size to pad.
        pad_value (Any): Filled value for padding. Defaults to `None`.

    Returns:
        Tensor: Padded tensor.
    """
    return __pad_with_value_if_necessary(
        x, pad_dim, pad_size=pad_size, pad_value=pad_value
    )


def __pad_with_value_if_necessary(
    x: Tensor, pad_dim: int, pad_size: int, pad_value: Optional[Any] = None
) -> Tensor:
    """Pad a tensor with a value along some dim, do nothing on default.

    Args:
        x (Tensor): Input tensor.
        pad_dim (int): Along which dim to pad.
        pad_size (int): To which size to pad.
        pad_value (Any): Filled value for padding. Defaults to `None`.

    Returns:
        Tensor: Padded tensor.
    """
    return x


@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.codebase.mmdet.deploy.utils.__pad_with_value_if_necessary",
    backend=Backend.TENSORRT.value,
)
def __pad_with_value_if_necessary__tensorrt(
    x: Tensor, pad_dim: int, pad_size: int, pad_value: Optional[Any] = None
) -> Tensor:
    """Pad a tensor with a value along some dim.

    Args:
        x (Tensor): Input tensor.
        pad_dim (int): Along which dim to pad.
        pad_size (int): To which size to pad.
        pad_value (Any): Filled value for padding. Defaults to `None`.

    Returns:
        Tensor: Padded tensor.
    """
    return pad_with_value(x, pad_dim, pad_size=pad_size, pad_value=pad_value)


def __gather_topk(
    *inputs: Sequence[torch.Tensor],
    inds: torch.Tensor,
    batch_size: int,
    is_batched: bool = True,
) -> Tuple[torch.Tensor]:
    """The default implementation of gather_topk."""
    if is_batched:
        batch_inds = torch.arange(batch_size, device=inds.device).unsqueeze(-1)
        outputs = [x[batch_inds, inds, ...] if x is not None else None for x in inputs]
    else:
        prior_inds = inds.new_zeros((1, 1))
        outputs = [x[prior_inds, inds, ...] if x is not None else None for x in inputs]

    return outputs


class TRTGatherTopk(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x: torch.Tensor, inds: torch.Tensor):
        """Implement of gather topk."""
        batch_size = x.size(0)
        batch_inds = torch.arange(batch_size, device=inds.device).unsqueeze(-1)
        return x[batch_inds, inds, ...]

    @staticmethod
    def symbolic(g, x, inds):
        """symbolic of gather topk."""
        out = g.op("mmdeploy::GatherTopk", x, inds, outputs=1)

        return out


@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.codebase.mmdet.deploy.utils.__gather_topk", backend=Backend.TENSORRT.value
)
def __gather_topk__trt(
    *inputs: Sequence[torch.Tensor],
    inds: torch.Tensor,
    batch_size: int,
    is_batched: bool = True,
) -> Tuple[torch.Tensor]:
    """TensorRT gather_topk."""
    ctx = FUNCTION_REWRITER.get_context()
    _ = ctx
    if is_batched:
        index_shape = inds.shape
        index_dim = inds.dim()
        outputs = [None for _ in inputs]
        for i, x in enumerate(inputs):
            if x is None:
                continue
            out = TRTGatherTopk.apply(x, inds).to(x.dtype)
            out_shape = [*index_shape, *x.shape[index_dim:]]
            out = out.reshape(out_shape)
            outputs[i] = out
    else:
        prior_inds = inds.new_zeros((1, 1))
        outputs = [x[prior_inds, inds, ...] if x is not None else None for x in inputs]

    return outputs


@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.codebase.mmdet.deploy.utils.__gather_topk", backend=Backend.ASCEND.value
)
@FUNCTION_REWRITER.register_rewriter(
    "mmdeploy.codebase.mmdet.deploy.utils.__gather_topk", backend=Backend.COREML.value
)
def __gather_topk__nonbatch(
    *inputs: Sequence[torch.Tensor],
    inds: torch.Tensor,
    batch_size: int,
    is_batched: bool = True,
) -> Tuple[torch.Tensor]:
    """Single batch gather_topk."""
    assert batch_size == 1
    inds = inds.squeeze(0)
    outputs = [x[:, inds, ...] if x is not None else None for x in inputs]

    return outputs


def gather_topk(
    *inputs: Sequence[torch.Tensor],
    inds: torch.Tensor,
    batch_size: int,
    is_batched: bool = True,
) -> Tuple[torch.Tensor]:
    """Gather topk of each tensor.

    Args:
        inputs (Sequence[torch.Tensor]): Tensors to be gathered.
        inds (torch.Tensor): Topk index.
        batch_size (int): batch_size.
        is_batched (bool): Inputs is batched or not.

    Returns:
        Tuple[torch.Tensor]: Gathered tensors.
    """
    import mmdeploy

    outputs = juxtapose.mmdeploy.codebase.mmdet.deploy.utils.__gather_topk(
        *inputs, inds=inds, batch_size=batch_size, is_batched=is_batched
    )

    if len(outputs) == 1:
        outputs = outputs[0]
    return outputs



# Copyright (c) OpenMMLab. All rights reserved.
from .calibration import create_calib_input_data
from .utils import (build_task_processor, get_predefined_partition_cfg,
                    to_backend)

__all__ = [
    'create_calib_input_data', 'build_task_processor',
    'get_predefined_partition_cfg', 'to_backend'
]



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union

import mmcv
import numpy as np
import torch
from mmengine import Config
from mmengine.model import BaseDataPreprocessor
from torch.utils.data import DataLoader, Dataset

from juxtapose.mmdeploy.utils import (
    get_backend_config,
    get_codebase,
    get_codebase_config,
    get_root_logger,
)
from juxtapose.mmdeploy.utils.config_utils import (
    get_codebase_external_module,
    get_rknn_quantization,
)
from juxtapose.mmdeploy.utils.dataset import is_can_sort_dataset, sort_dataset


class BaseTask(metaclass=ABCMeta):
    """Wrap the processing functions of a Computer Vision task.

    Args:
        model_cfg (str | Config): Model config file.
        deploy_cfg (str | Config): Deployment config file.
        device (str): A string specifying device type.
        experiment_name (str, optional): Name of current experiment.
            If not specified, timestamp will be used as
            ``experiment_name``. Defaults to ``None``.
    """

    def __init__(
        self,
        model_cfg: Config,
        deploy_cfg: Config,
        device: str,
        experiment_name: str = "BaseTask",
    ):
        self.model_cfg = model_cfg
        self.deploy_cfg = deploy_cfg
        self.device = device

        self.codebase = get_codebase(deploy_cfg)
        self.experiment_name = experiment_name

        # init scope
        from .. import import_codebase

        custom_module_list = get_codebase_external_module(deploy_cfg)
        import_codebase(self.codebase, custom_module_list)

        from mmengine.registry import DefaultScope

        if not DefaultScope.check_instance_created(self.experiment_name):
            self.scope = DefaultScope.get_instance(
                self.experiment_name, scope_name=self.model_cfg.get("default_scope")
            )
        else:
            self.scope = DefaultScope.get_instance(self.experiment_name)

        # lazy build visualizer
        self.visualizer = self.model_cfg.get("visualizer", None)

    @abstractmethod
    def build_backend_model(
        self,
        model_files: Sequence[str] = None,
        data_preprocessor_updater: Optional[Callable] = None,
        **kwargs,
    ) -> torch.nn.Module:
        """Initialize backend model.

        Args:
            model_files (Sequence[str]): Input model files.
            data_preprocessor_updater (Callable | None): A function to update
                the data_preprocessor. Defaults to None.

        Returns:
            nn.Module: An initialized backend model.
        """
        pass

    def build_data_preprocessor(self):
        """build data preprocessor.

        Returns:
            BaseDataPreprocessor:
                Initialized instance of :class:`BaseDataPreprocessor`.
        """
        model = deepcopy(self.model_cfg.model)
        preprocess_cfg = model["data_preprocessor"]

        from mmengine.registry import MODELS

        data_preprocessor = MODELS.build(preprocess_cfg)
        data_preprocessor.to(self.device)

        return data_preprocessor

    def build_pytorch_model(
        self,
        model_checkpoint: Optional[str] = None,
        cfg_options: Optional[Dict] = None,
        **kwargs,
    ) -> torch.nn.Module:
        """Initialize torch model.

        Args:
            model_checkpoint (str): The checkpoint file of torch model,
                defaults to `None`.
            cfg_options (dict): Optional config key-pair parameters.

        Returns:
            nn.Module: An initialized torch model generated by other OpenMMLab
                codebases.
        """
        from mmengine.model import revert_sync_batchnorm
        from mmengine.registry import MODELS

        model = deepcopy(self.model_cfg.model)
        model.pop("pretrained", None)
        preprocess_cfg = deepcopy(self.model_cfg.get("preprocess_cfg", {}))
        preprocess_cfg.update(deepcopy(self.model_cfg.get("data_preprocessor", {})))
        model.setdefault("data_preprocessor", preprocess_cfg)
        model = MODELS.build(model)
        if model_checkpoint is not None:
            from mmengine.runner.checkpoint import load_checkpoint

            load_checkpoint(model, model_checkpoint, map_location=self.device)

        model = revert_sync_batchnorm(model)
        if hasattr(model, "backbone") and hasattr(model.backbone, "switch_to_deploy"):
            model.backbone.switch_to_deploy()
        model = model.to(self.device)
        model.eval()
        return model

    def build_dataset(
        self, dataset_cfg: Union[str, Config], is_sort_dataset: bool = False, **kwargs
    ) -> Dataset:
        """Build dataset for different codebase.

        Args:
            dataset_cfg (str | Config): Dataset config file or Config
                object.
            is_sort_dataset (bool): When 'True', the dataset will be sorted
                by image shape in ascending order if 'dataset_cfg'
                contains information about height and width.
                Default is `False`.

        Returns:
            Dataset: The built dataset.
        """
        backend_cfg = get_backend_config(self.deploy_cfg)
        from juxtapose.mmdeploy.utils import load_config

        dataset_cfg = load_config(dataset_cfg)[0]
        if "pipeline" in backend_cfg:
            dataset_cfg.pipeline = backend_cfg.pipeline
        from mmengine.registry import DATASETS

        dataset = DATASETS.build(dataset_cfg)
        logger = get_root_logger()
        if is_sort_dataset:
            if is_can_sort_dataset(dataset):
                sort_dataset(dataset)
            else:
                logger.info(
                    "Sorting the dataset by 'height' and 'width' " "is not possible."
                )
        return dataset

    @staticmethod
    def build_dataloader(
        dataloader: Union[DataLoader, Dict], seed: Optional[int] = None
    ) -> DataLoader:
        """Build PyTorch dataloader. A wrap of Runner.build_dataloader.

        Args:
            dataloader (DataLoader or dict): A Dataloader object or a dict to
                build Dataloader object. If ``dataloader`` is a Dataloader
                object, just returns itself.
            seed (int, optional): Random seed. Defaults to None.

        Returns:
            Dataloader: DataLoader build from ``dataloader_cfg``.
        """
        from mmengine.runner import Runner

        return Runner.build_dataloader(dataloader, seed)

    def build_test_runner(
        self,
        model: torch.nn.Module,
        work_dir: str,
        log_file: Optional[str] = None,
        show: bool = False,
        show_dir: Optional[str] = None,
        wait_time: int = 0,
        interval: int = 1,
        dataloader: Optional[Union[DataLoader, Dict]] = None,
    ):
        def _merge_cfg(cfg):
            """Merge CLI arguments to config."""
            # -------------------- visualization --------------------
            if show or (show_dir is not None):
                assert "visualization" in cfg.default_hooks, (
                    "VisualizationHook is not set in the `default_hooks`"
                    " field of config. Please set "
                    '`visualization=dict(type="VisualizationHook")`'
                )

                cfg.default_hooks.visualization.draw = True
                cfg.default_hooks.visualization.show = show
                cfg.default_hooks.visualization.wait_time = wait_time
                cfg.default_hooks.visualization.test_out_dir = show_dir
                cfg.default_hooks.visualization.interval = interval

            return cfg

        model_cfg = deepcopy(self.model_cfg)
        if dataloader is None:
            dataloader = model_cfg.test_dataloader
        if not isinstance(dataloader, DataLoader):
            if type(dataloader) == list:
                dataloader = [self.build_dataloader(dl) for dl in dataloader]
            else:
                dataloader = self.build_dataloader(dataloader)

        model_cfg = _merge_cfg(model_cfg)

        visualizer = self.get_visualizer(work_dir, work_dir)
        from .runner import DeployTestRunner

        runner = DeployTestRunner(
            model=model,
            work_dir=work_dir,
            log_file=log_file,
            device=self.device,
            visualizer=visualizer,
            default_hooks=model_cfg.default_hooks,
            test_dataloader=dataloader,
            test_cfg=model_cfg.test_cfg,
            test_evaluator=model_cfg.test_evaluator,
            default_scope=model_cfg.default_scope,
        )
        return runner

    def update_data_preprocessor(self, data_preprocessor: Config):
        """Update data_preprocessor.

        Args:
            data_preprocessor (mmengine.Config): The data preprocessor.
        Returns:
            mmengine.Config: The updated data preprocessor.
        """
        data_preprocessor = deepcopy(data_preprocessor)
        if get_rknn_quantization(self.deploy_cfg):
            if data_preprocessor is not None:
                data_preprocessor["mean"] = [0, 0, 0]
                data_preprocessor["std"] = [1, 1, 1]
        return data_preprocessor

    @abstractmethod
    def create_input(
        self,
        imgs: Union[str, np.ndarray],
        input_shape: Optional[Sequence[int]] = None,
        data_preprocessor: Optional[BaseDataPreprocessor] = None,
        **kwargs,
    ) -> Tuple[Dict, torch.Tensor]:
        """Create input for model.

        Args:
            imgs (str | np.ndarray | Sequence): Input image(s),
                accepted data types are `str`, `np.ndarray`.
            input_shape (list[int]): Input shape of image in (width, height)
                format, defaults to `None`.

        Returns:
            tuple: (data, img), meta information for the input image and input
                image tensor.
        """
        pass

    def get_visualizer(self, name: str, save_dir: str):
        """Get the visualizer instance.

        Args:
            name (str): The name of the visualizer.
            save_dir (str): The save directory of visualizer.
        """
        cfg = deepcopy(self.visualizer)
        cfg.name = name
        cfg.save_dir = save_dir
        from mmengine.registry import VISUALIZERS, DefaultScope

        with DefaultScope.overwrite_default_scope(cfg.pop("_scope_", None)):
            # get the global default scope
            default_scope = DefaultScope.get_current_instance()
            if default_scope is not None:
                scope_name = default_scope.scope_name
                root = VISUALIZERS._get_root_registry()
                registry = root._search_child(scope_name)
                if registry is None:
                    registry = VISUALIZERS
            else:
                registry = VISUALIZERS
            VisualizerClass = registry.get(cfg.type)
            if VisualizerClass.check_instance_created(cfg.name):
                return VisualizerClass.get_instance(cfg.name)
            else:
                return registry.build_func(cfg, registry=registry)

    def visualize(
        self,
        image: Union[str, np.ndarray],
        result: list,
        output_file: str,
        window_name: str = "",
        show_result: bool = False,
        draw_gt: bool = False,
        **kwargs,
    ):
        """Visualize predictions of a model.

        Args:
            model (nn.Module): Input model.
            image (str | np.ndarray): Input image to draw predictions on.
            result (list): A list of predictions.
            output_file (str): Output file to save drawn image.
            window_name (str): The name of visualization window. Defaults to
                an empty string.
            show_result (bool): Whether to show result in windows, defaults
                to `False`.
            draw_gt (bool): Whether to show ground truth in windows, defaults
                to `False`.
        """
        save_dir, save_name = osp.split(output_file)
        visualizer = self.get_visualizer(window_name, save_dir)

        name = osp.splitext(save_name)[0]
        if isinstance(image, str):
            image = mmcv.imread(image, channel_order="rgb")
        assert isinstance(image, np.ndarray)

        visualizer.add_datasample(
            name,
            image,
            data_sample=result,
            draw_gt=draw_gt,
            show=show_result,
            out_file=output_file,
        )

    @staticmethod
    @abstractmethod
    def get_partition_cfg(partition_type: str, **kwargs) -> Dict:
        """Get a certain partition config.

        Args:
            partition_type (str): A string specifying partition type.

        Returns:
            dict: A dictionary of partition config.
        """
        pass

    @staticmethod
    def get_tensor_from_input(input_data: Dict[str, Any], **kwargs) -> torch.Tensor:
        """Get input tensor from input data.

        Args:
            input_data (dict): Input data containing meta info and image
                tensor.
        Returns:
            torch.Tensor: An image in `Tensor`.
        """
        return input_data["inputs"]

    @abstractmethod
    def get_preprocess(self, *args, **kwargs) -> Dict:
        """Get the preprocess information for SDK.

        Return:
            dict: Composed of the preprocess information.
        """
        pass

    @abstractmethod
    def get_postprocess(self, *args, **kwargs) -> Dict:
        """Get the postprocess information for SDK.

        Return:
            dict: Composed of the postprocess information.
        """
        pass

    @abstractmethod
    def get_model_name(self, *args, **kwargs) -> str:
        """Get the model name.

        Return:
            str: the name of the model.
        """
        pass

    @property
    def from_mmrazor(self) -> bool:
        """Whether the codebase from mmrazor.

        Returns:
            bool: From mmrazor or not.

        Raises:
            TypeError: An error when type of `from_mmrazor` is not boolean.
        """
        codebase_config = get_codebase_config(self.deploy_cfg)
        from_mmrazor = codebase_config.get("from_mmrazor", False)
        if not isinstance(from_mmrazor, bool):
            raise TypeError(
                "`from_mmrazor` attribute must be boolean type! "
                f"but got: {from_mmrazor}"
            )

        return from_mmrazor



# Copyright (c) OpenMMLab. All rights reserved.
import torch
from packaging import version
from torch import Tensor
from torch.onnx import symbolic_helper as sym_help

from juxtapose.mmdeploy.core import FUNCTION_REWRITER, mark
from juxtapose.mmdeploy.utils import IR, is_dynamic_batch
from juxtapose.mmdeploy.utils.constants import Backend
from .nms_rotated import multiclass_nms_rotated


class ONNXNMSop(torch.autograd.Function):
    """Create onnx::NonMaxSuppression op.

    NMS in mmcv only supports one class with no batch info. This class assists
    in exporting NMS of ONNX's definition.
    """

    @staticmethod
    def forward(
        ctx,
        boxes: Tensor,
        scores: Tensor,
        max_output_boxes_per_class: int,
        iou_threshold: float,
        score_threshold: float,
    ) -> Tensor:
        """Get NMS output indices.

        Args:
            ctx (Context): The context with meta information.
            boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            max_output_boxes_per_class (int): Maximum number of output
                boxes per class of nms.
            iou_threshold (float): IOU threshold of nms.
            score_threshold (float): score threshold of nms.

        Returns:
            Tensor: Selected indices of boxes. 2-D tensor of shape
            (num_selected_indices, 3) with each row of
            [batch_index, class_index, box_index].
        """
        from mmcv.ops import nms

        batch_size, num_class, _ = scores.shape

        score_threshold = float(score_threshold)
        iou_threshold = float(iou_threshold)
        indices = []
        for batch_id in range(batch_size):
            for cls_id in range(num_class):
                _boxes = boxes[batch_id, ...]
                # score_threshold=0 requires scores to be contiguous
                _scores = scores[batch_id, cls_id, ...].contiguous()
                _, box_inds = nms(
                    _boxes,
                    _scores,
                    iou_threshold,
                    offset=0,
                    score_threshold=score_threshold,
                    max_num=max_output_boxes_per_class,
                )
                batch_inds = torch.zeros_like(box_inds) + batch_id
                cls_inds = torch.zeros_like(box_inds) + cls_id
                indices.append(torch.stack([batch_inds, cls_inds, box_inds], dim=-1))
        indices = torch.cat(indices)
        return indices

    @staticmethod
    def symbolic(
        g,
        boxes: Tensor,
        scores: Tensor,
        max_output_boxes_per_class: int,
        iou_threshold: float,
        score_threshold: float,
    ):
        """Symbolic function for onnx::NonMaxSuppression.

        Args:
            g (Graph): The traced onnx graph.
            boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            max_output_boxes_per_class (int): Maximum number of output
                boxes per class of nms.
            iou_threshold (float): IOU threshold of nms.
            score_threshold (float): score threshold of nms.

        Returns:
            NonMaxSuppression op for onnx.
        """
        if not sym_help._is_value(max_output_boxes_per_class):
            max_output_boxes_per_class = g.op(
                "Constant",
                value_t=torch.tensor(max_output_boxes_per_class, dtype=torch.long),
            )

        if not sym_help._is_value(iou_threshold):
            iou_threshold = g.op(
                "Constant", value_t=torch.tensor([iou_threshold], dtype=torch.float)
            )

        if not sym_help._is_value(score_threshold):
            score_threshold = g.op(
                "Constant", value_t=torch.tensor([score_threshold], dtype=torch.float)
            )
        return g.op(
            "NonMaxSuppression",
            boxes,
            scores,
            max_output_boxes_per_class,
            iou_threshold,
            score_threshold,
        )


class TRTBatchedNMSop(torch.autograd.Function):
    """Create mmdeploy::TRTBatchedNMS op for TensorRT backend.

    NMS in ONNX supports dynamic outputs. This class helps replace
    onnx::NonMaxSuppression with mmdeploy::TRTBatchedNMS.
    """

    @staticmethod
    def forward(
        ctx,
        boxes: Tensor,
        scores: Tensor,
        num_classes: int,
        pre_topk: int,
        after_topk: int,
        iou_threshold: float,
        score_threshold: float,
        background_label_id: int = -1,
        return_index: bool = False,
    ):
        """Forward of batched nms.

        Args:
            ctx (Context): The context with meta information.
            boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
            scores (Tensor): The detection scores of shape
                [N, num_boxes, num_classes].
            num_classes (int): MThe number of classes in the network.
            pre_topk (int): The number of bounding boxes to be fed into
                the NMS step.
            after_topk (int): The number of total bounding boxes to be kept
                per-image after the NMS step. Should be less than or equal
                to the pre_topk value.
            iou_threshold (float): IOU threshold of nms.
            score_threshold (float): score threshold of nms.
            background_label_id (int): The label ID for the background class.
                If there is no background class, set it to -1.

        Returns:
            Tensor: Selected indices of boxes. 2-D tensor of shape
            (num_selected_indices, 3) with each row of
            [batch_index, class_index, box_index]. Note it is generated
            randomly to make it exportable to onnx.
        """
        batch_size, num_boxes, num_classes = scores.shape

        out_boxes = min(num_boxes, after_topk)
        ret = (
            torch.rand(batch_size, out_boxes, 5).to(scores.device),
            torch.randint(0, num_classes, (batch_size, out_boxes)).to(scores.device),
        )
        if return_index:
            ret = ret + (
                torch.randint(0, out_boxes, (batch_size, out_boxes)).to(scores.device),
            )
        return ret

    @staticmethod
    def symbolic(
        g,
        boxes: Tensor,
        scores: Tensor,
        num_classes: int,
        pre_topk: int,
        after_topk: int,
        iou_threshold: float,
        score_threshold: float,
        background_label_id: int = -1,
        return_index: bool = False,
    ):
        """Symbolic function for mmdeploy::TRTBatchedNMS."""
        return g.op(
            "mmdeploy::TRTBatchedNMS",
            boxes,
            scores,
            num_classes_i=num_classes,
            background_label_id_i=background_label_id,
            iou_threshold_f=iou_threshold,
            score_threshold_f=score_threshold,
            topk_i=pre_topk,
            keep_topk_i=after_topk,
            is_normalized_i=False,
            clip_boxes_i=False,
            return_index_i=return_index,
            outputs=3 if return_index else 2,
        )


def _select_nms_index(
    scores: torch.Tensor,
    boxes: torch.Tensor,
    nms_index: torch.Tensor,
    batch_size: int,
    keep_top_k: int = -1,
    pre_inds: torch.Tensor = None,
    output_index: bool = False,
):
    """Transform NMS output.

    Args:
        scores (Tensor): The detection scores of shape
            [N, num_classes, num_boxes].
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
        nms_index (Tensor): NMS output of bounding boxes indexing.
        batch_size (int): Batch size of the input image.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.
        pre_inds (Tensor): The pre-topk indices of boxes before nms.
            Defaults to None.
        return_index (bool): Whether to return indices of original bboxes.
            Defaults to False.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 5]
            and `labels` of shape [N, num_det].
    """
    batch_inds, cls_inds = nms_index[:, 0], nms_index[:, 1]
    box_inds = nms_index[:, 2]

    # index by nms output
    scores = scores[batch_inds, cls_inds, box_inds].unsqueeze(1)
    boxes = boxes[batch_inds, box_inds, ...]
    dets = torch.cat([boxes, scores], dim=1)

    # batch all
    batched_dets = dets.unsqueeze(0).repeat(batch_size, 1, 1)
    batch_template = torch.arange(
        0, batch_size, dtype=batch_inds.dtype, device=batch_inds.device
    )
    batched_dets = batched_dets.where(
        (batch_inds == batch_template.unsqueeze(1)).unsqueeze(-1),
        batched_dets.new_zeros(1),
    )

    batched_labels = cls_inds.unsqueeze(0).repeat(batch_size, 1)
    batched_labels = batched_labels.where(
        (batch_inds == batch_template.unsqueeze(1)), batched_labels.new_ones(1) * -1
    )

    N = batched_dets.shape[0]

    # expand tensor to eliminate [0, ...] tensor
    batched_dets = torch.cat((batched_dets, batched_dets.new_zeros((N, 1, 5))), 1)
    batched_labels = torch.cat((batched_labels, batched_labels.new_zeros((N, 1))), 1)
    if output_index and pre_inds is not None:
        # batch all
        pre_inds = pre_inds[batch_inds, box_inds]
        pre_inds = pre_inds.unsqueeze(0).repeat(batch_size, 1)
        pre_inds = pre_inds.where(
            (batch_inds == batch_template.unsqueeze(1)), pre_inds.new_zeros(1)
        )
        pre_inds = torch.cat((pre_inds, pre_inds.new_zeros((N, 1))), 1)
    # sort
    is_use_topk = keep_top_k > 0 and (
        torch.onnx.is_in_onnx_export() or keep_top_k < batched_dets.shape[1]
    )
    if is_use_topk:
        _, topk_inds = batched_dets[:, :, -1].topk(keep_top_k, dim=1)
    else:
        _, topk_inds = batched_dets[:, :, -1].sort(dim=1, descending=True)
    topk_batch_inds = torch.arange(
        batch_size, dtype=topk_inds.dtype, device=topk_inds.device
    ).view(-1, 1)
    batched_dets = batched_dets[topk_batch_inds, topk_inds, ...]
    batched_labels = batched_labels[topk_batch_inds, topk_inds, ...]
    if output_index:
        if pre_inds is not None:
            topk_inds = pre_inds[topk_batch_inds, topk_inds, ...]
        topk_inds = topk_inds[:, :-1]
        return batched_dets, batched_labels, topk_inds
    # slice and recover the tensor
    return batched_dets, batched_labels


def _multiclass_nms(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
):
    """Create a dummy onnx::NonMaxSuppression op while exporting to ONNX.

    This function helps exporting to onnx with batch and multiclass NMS op. It
    only supports class-agnostic detection results. That is, the scores is of
    shape (N, num_bboxes, num_classes) and the boxes is of shape (N, num_boxes,
    4).
    """
    if version.parse(torch.__version__) < version.parse("1.13.0"):
        max_output_boxes_per_class = torch.LongTensor([max_output_boxes_per_class])
    iou_threshold = torch.tensor([iou_threshold], dtype=torch.float32)
    score_threshold = torch.tensor([score_threshold], dtype=torch.float32)
    batch_size = scores.shape[0]

    if pre_top_k > 0:
        max_scores, _ = scores.max(-1)
        _, topk_inds = max_scores.topk(pre_top_k)
        batch_inds = torch.arange(batch_size, device=scores.device).view(-1, 1).long()
        boxes = boxes[batch_inds, topk_inds, :]
        scores = scores[batch_inds, topk_inds, :]

    scores = scores.permute(0, 2, 1)
    selected_indices = ONNXNMSop.apply(
        boxes, scores, max_output_boxes_per_class, iou_threshold, score_threshold
    )

    return _select_nms_index(
        scores,
        boxes,
        selected_indices,
        batch_size,
        keep_top_k=keep_top_k,
        pre_inds=topk_inds,
        output_index=output_index,
    )


def _multiclass_nms_single(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
):
    """Create a dummy onnx::NonMaxSuppression op while exporting to ONNX.

    Single batch nms could be optimized.
    """
    if version.parse(torch.__version__) < version.parse("1.13.0"):
        max_output_boxes_per_class = torch.LongTensor([max_output_boxes_per_class])
    iou_threshold = torch.tensor([iou_threshold], dtype=torch.float32)
    score_threshold = torch.tensor([score_threshold], dtype=torch.float32)

    # pre topk
    pre_topk_inds = None
    if pre_top_k > 0:
        max_scores, _ = scores.max(-1)
        _, topk_inds = max_scores.squeeze(0).topk(pre_top_k)
        boxes = boxes[:, topk_inds, :]
        scores = scores[:, topk_inds, :]
        pre_topk_inds = topk_inds

    scores = scores.permute(0, 2, 1)
    selected_indices = ONNXNMSop.apply(
        boxes, scores, max_output_boxes_per_class, iou_threshold, score_threshold
    )

    cls_inds = selected_indices[:, 1]
    box_inds = selected_indices[:, 2]

    scores = scores[:, cls_inds, box_inds].unsqueeze(2)
    boxes = boxes[:, box_inds, ...]
    dets = torch.cat([boxes, scores], dim=2)
    labels = cls_inds.unsqueeze(0)

    # pad
    dets = torch.cat((dets, dets.new_zeros((1, 1, 5))), 1)
    labels = torch.cat((labels, labels.new_zeros((1, 1))), 1)

    # topk or sort
    is_use_topk = keep_top_k > 0 and (
        torch.onnx.is_in_onnx_export() or keep_top_k < dets.shape[1]
    )
    if is_use_topk:
        _, topk_inds = dets[:, :, -1].topk(keep_top_k, dim=1)
    else:
        _, topk_inds = dets[:, :, -1].sort(dim=1, descending=True)
    topk_inds = topk_inds.squeeze(0)
    dets = dets[:, topk_inds, ...]
    labels = labels[:, topk_inds, ...]

    if output_index:
        bbox_index = box_inds.unsqueeze(0)
        if pre_top_k > 0:
            bbox_index = pre_topk_inds[None, box_inds]
        if keep_top_k > 0:
            bbox_index = bbox_index[:, topk_inds[:-1]]
        return dets, labels, bbox_index
    else:
        return dets, labels


@FUNCTION_REWRITER.register_rewriter(func_name="mmdeploy.mmcv.ops.nms._multiclass_nms")
def multiclass_nms__default(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
):
    """Create a dummy onnx::NonMaxSuppression op while exporting to ONNX.

    This function helps exporting to onnx with batch and multiclass NMS op.
    It only supports class-agnostic detection results. That is, the scores
    is of shape (N, num_bboxes, num_classes) and the boxes is of shape
    (N, num_boxes, 4).

    Args:
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        max_output_boxes_per_class (int): Maximum number of output
            boxes per class of nms. Defaults to 1000.
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): score threshold of nms.
            Defaults to 0.05.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 5]
            and `labels` of shape [N, num_det].
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg
    batch_size = boxes.size(0)
    if not is_dynamic_batch(deploy_cfg) and batch_size == 1:
        return _multiclass_nms_single(
            boxes,
            scores,
            max_output_boxes_per_class=max_output_boxes_per_class,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
            output_index=output_index,
        )
    else:
        return ctx.origin_func(
            boxes,
            scores,
            max_output_boxes_per_class=max_output_boxes_per_class,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
            output_index=output_index,
        )


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.mmcv.ops.nms._multiclass_nms", backend="tensorrt"
)
def multiclass_nms_static(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
):
    """Wrapper for `multiclass_nms` with TensorRT.

    Args:
        ctx (ContextCaller): The context with additional information.
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        max_output_boxes_per_class (int): Maximum number of output
            boxes per class of nms. Defaults to 1000.
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): score threshold of nms.
            Defaults to 0.05.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 5]
            and `labels` of shape [N, num_det].
    """
    boxes = boxes if boxes.dim() == 4 else boxes.unsqueeze(2)
    keep_top_k = (
        max_output_boxes_per_class
        if keep_top_k < 0
        else min(max_output_boxes_per_class, keep_top_k)
    )
    nms_output = TRTBatchedNMSop.apply(
        boxes,
        scores,
        int(scores.shape[-1]),
        pre_top_k,
        keep_top_k,
        iou_threshold,
        score_threshold,
        -1,
        output_index,
    )

    dets = nms_output[0]
    labels = nms_output[1]
    box_index = None if len(nms_output) <= 2 else nms_output[2]

    # retain shape info
    batch_size = boxes.size(0)

    dets_shape = dets.shape
    label_shape = labels.shape
    dets = dets.reshape([batch_size, *dets_shape[1:]])
    labels = labels.reshape([batch_size, *label_shape[1:]])
    if output_index:
        return dets, labels, box_index
    return dets, labels


@mark("multiclass_nms", inputs=["boxes", "scores"], outputs=["dets", "labels", "index"])
def multiclass_nms(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
    nms_type="nms",
):
    """Apis for multiclass nms."""
    if nms_type == "nms":
        return _multiclass_nms(
            boxes,
            scores,
            max_output_boxes_per_class=max_output_boxes_per_class,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
            output_index=output_index,
        )
    elif nms_type == "nms_rotated":
        return multiclass_nms_rotated(
            boxes,
            scores,
            max_output_boxes_per_class=max_output_boxes_per_class,
            iou_threshold=iou_threshold,
            score_threshold=score_threshold,
            pre_top_k=pre_top_k,
            keep_top_k=keep_top_k,
        )
    else:
        raise NotImplementedError(f"Unsupported nms type: {nms_type}.")


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.mmcv.ops.nms._multiclass_nms", backend=Backend.COREML.value
)
def multiclass_nms__coreml(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
):
    """rewrite for coreml batched nms.

    Use coreml_nms from custom ops.
    """

    # load custom nms
    from juxtapose.mmdeploy.backend.torchscript import get_ops_path, ops_available

    assert ops_available(), "coreml require custom torchscript ops support."
    torch.ops.load_library(get_ops_path())
    try:
        coreml_nms = torch.ops.mmdeploy.coreml_nms
    except Exception:
        raise Exception("Can not use coreml_nms. Please build torchscript custom ops.")

    batch_size = scores.shape[0]
    assert batch_size == 1, "batched nms is not supported for now."

    # pre-topk
    if pre_top_k > 0:
        max_scores, _ = scores.max(-1)
        _, topk_inds = max_scores.topk(pre_top_k)
        boxes = boxes[:, topk_inds.squeeze(), ...]
        scores = scores[:, topk_inds.squeeze(), ...]

    def _xyxy2xywh(boxes):
        xy0 = boxes[..., :2]
        xy1 = boxes[..., 2:]
        xy = (xy0 + xy1) / 2
        wh = xy1 - xy0
        return torch.cat([xy, wh], dim=-1)

    def _xywh2xyxy(boxes):
        xy = boxes[..., :2]
        half_wh = boxes[..., 2:] / 2
        return torch.cat([xy - half_wh, xy + half_wh], dim=-1)

    boxes = _xyxy2xywh(boxes)
    keep_top_k = keep_top_k if keep_top_k > 0 else max_output_boxes_per_class
    boxes, scores, box_index, _ = coreml_nms(
        boxes,
        scores,
        iou_threshold,
        score_threshold,
        min(keep_top_k, max_output_boxes_per_class),
    )

    scores, labels = scores.max(-1)
    boxes = _xywh2xyxy(boxes)
    dets = torch.cat([boxes, scores.unsqueeze(-1)], dim=-1)

    if output_index:
        return dets, labels, box_index
    return dets, labels


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.mmcv.ops.nms._multiclass_nms", ir=IR.TORCHSCRIPT
)
def multiclass_nms__torchscript(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index=False,
):
    """rewrite for torchscript batched nms.

    Use batched_nms from torchvision instead of custom nms.
    """
    assert not output_index, "output_index is not supported on this backend."
    # TODO: simplify inference for non-batch model
    from torchvision.ops import batched_nms

    batch_size = scores.shape[0]
    num_boxes = scores.shape[1]
    num_classes = scores.shape[2]
    box_per_cls = len(boxes.shape) == 4
    scores = torch.where(scores > score_threshold, scores, scores.new_zeros(1))

    # pre-topk
    if pre_top_k > 0:
        max_scores, _ = scores.max(-1)
        _, topk_inds = max_scores.topk(pre_top_k)
        batch_inds = torch.arange(batch_size).view(-1, 1).long()
        boxes = boxes[batch_inds, topk_inds, ...]
        scores = scores[batch_inds, topk_inds, :]
        num_boxes = scores.shape[1]

    idxs = torch.arange(0, batch_size, device=scores.device).unsqueeze(1)
    idxs = idxs.repeat(1, num_boxes).view(-1)

    keeps = [None] * num_classes
    for cls_id in range(num_classes):
        box = boxes if not box_per_cls else boxes[:, :, cls_id, :]
        score = scores[:, :, cls_id]
        box = box.view(-1, 4)
        score = score.view(-1)
        box_keep = batched_nms(box, score, idxs, iou_threshold=iou_threshold)
        box_keep = box_keep[: max_output_boxes_per_class * batch_size]
        batch_keep = idxs[box_keep]
        cls_keep = torch.ones_like(box_keep) * cls_id
        box_keep = box_keep - batch_keep * num_boxes
        keeps[cls_id] = torch.stack([batch_keep, cls_keep, box_keep], dim=1)

    keeps = torch.cat(keeps)
    scores = scores.permute(0, 2, 1)
    dets, labels = _select_nms_index(
        scores, boxes, keeps, batch_size, keep_top_k=keep_top_k
    )

    return dets, labels


class AscendBatchNMSOp(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        bboxes: torch.Tensor,
        scores: torch.Tensor,
        score_threshold: float,
        iou_threshold: float,
        max_size_per_class: int,
        max_total_size: int,
    ):
        """Dummy nms forward
        Args:
        boxes (torch.Tensor): boxes in shape (batch, N, C, 4).
        scores (torch.Tensor): scores in shape (batch, N, C).
        score_threshold (float): the score threshold.
        iou_threshold (float): the iou threshold.
        max_size_per_class (int): max size per class.
        max_total_size (int): max total size.

        Returns:
            (torch.Tensor): boxes,(1, N, 4)
            (torch.Tensor): scores,(1, N)
            (torch.Tensor): classes,(1, N)
            (torch.Tensor): num_dets,(1,)
        """

        # Python implementation for onnx export
        nmsed_boxes = bboxes[:, :max_total_size, 0, :]
        nmsed_scores = scores[:, :max_total_size, 0]
        nmsed_classes = torch.arange(max_total_size, dtype=torch.long)
        nmsed_num = torch.Tensor([max_total_size])

        return nmsed_boxes, nmsed_scores, nmsed_classes, nmsed_num

    @staticmethod
    def symbolic(g, bboxes, scores, score_thr, iou_thr, max_size_p_class, max_t_size):
        nmsed_boxes, nmsed_scores, nmsed_classes, nmsed_num = g.op(
            "mmdeploy::BatchMultiClassNMS",
            bboxes,
            scores,
            score_threshold_f=score_thr,
            iou_threshold_f=iou_thr,
            max_size_per_class_i=max_size_p_class,
            max_total_size_i=max_t_size,
            outputs=4,
        )
        return nmsed_boxes, nmsed_scores, nmsed_classes, nmsed_num


@FUNCTION_REWRITER.register_rewriter(
    func_name="mmdeploy.mmcv.ops.nms._multiclass_nms", backend="ascend"
)
def multiclass_nms__ascend(
    boxes: Tensor,
    scores: Tensor,
    max_output_boxes_per_class: int = 1000,
    iou_threshold: float = 0.5,
    score_threshold: float = 0.05,
    pre_top_k: int = -1,
    keep_top_k: int = -1,
    output_index: bool = False,
):
    """Wrapper for `multiclass_nms` with Ascend.

    Args:
        ctx (ContextCaller): The context with additional information.
        boxes (Tensor): The bounding boxes of shape [N, num_boxes, 4].
        scores (Tensor): The detection scores of shape
            [N, num_boxes, num_classes].
        max_output_boxes_per_class (int): Maximum number of output
            boxes per class of nms. Defaults to 1000.
        iou_threshold (float): IOU threshold of nms. Defaults to 0.5.
        score_threshold (float): score threshold of nms.
            Defaults to 0.05.
        pre_top_k (int): Number of top K boxes to keep before nms.
            Defaults to -1.
        keep_top_k (int): Number of top K boxes to keep after nms.
            Defaults to -1.

    Returns:
        tuple[Tensor, Tensor]: (dets, labels), `dets` of shape [N, num_det, 5]
            and `labels` of shape [N, num_det].
    """
    assert not output_index, "output_index is not supported on this backend."
    boxes = boxes if boxes.dim() == 4 else boxes.unsqueeze(2)
    keep_top_k = (
        max_output_boxes_per_class
        if keep_top_k < 0
        else min(max_output_boxes_per_class, keep_top_k)
    )
    nmsed_boxes, nmsed_scores, nmsed_classes, _ = AscendBatchNMSOp.apply(
        boxes, scores, score_threshold, iou_threshold, keep_top_k, keep_top_k
    )

    dets = torch.cat([nmsed_boxes, nmsed_scores.unsqueeze(2)], dim=-1)
    return dets, nmsed_classes.int()



# Copyright (c) OpenMMLab. All rights reserved.

from typing import Dict, Iterable, Optional, Union

import onnx

from .core import PIPELINE_MANAGER


@PIPELINE_MANAGER.register_pipeline()
def extract_model(
    model: Union[str, onnx.ModelProto],
    start_marker: Union[str, Iterable[str]],
    end_marker: Union[str, Iterable[str]],
    start_name_map: Optional[Dict[str, str]] = None,
    end_name_map: Optional[Dict[str, str]] = None,
    dynamic_axes: Optional[Dict[str, Dict[int, str]]] = None,
    save_file: Optional[str] = None,
) -> onnx.ModelProto:
    """Extract partition-model from an ONNX model.

    The partition-model is defined by the names of the input and output tensors
    exactly.

    Examples:
        >>> from juxtapose.mmdeploy.apis import extract_model
        >>> model = 'work_dir/fastrcnn.onnx'
        >>> start_marker = 'detector:input'
        >>> end_marker = ['extract_feat:output', 'multiclass_nms[0]:input']
        >>> dynamic_axes = {
            'input': {
                0: 'batch',
                2: 'height',
                3: 'width'
            },
            'scores': {
                0: 'batch',
                1: 'num_boxes',
            },
            'boxes': {
                0: 'batch',
                1: 'num_boxes',
            }
        }
        >>> save_file = 'partition_model.onnx'
        >>> extract_model(model, start_marker, end_marker, \
                dynamic_axes=dynamic_axes, \
                save_file=save_file)

    Args:
        model (str | onnx.ModelProto): Input ONNX model to be extracted.
        start_marker (str | Sequence[str]): Start marker(s) to extract.
        end_marker (str | Sequence[str]): End marker(s) to extract.
        start_name_map (Dict[str, str]): A mapping of start names, defaults to
            `None`.
        end_name_map (Dict[str, str]): A mapping of end names, defaults to
            `None`.
        dynamic_axes (Dict[str, Dict[int, str]]): A dictionary to specify
            dynamic axes of input/output, defaults to `None`.
        save_file (str): A file to save the extracted model, defaults to
            `None`.

    Returns:
        onnx.ModelProto: The extracted model.
    """

    from .onnx import extract_partition

    return extract_partition(
        model,
        start_marker,
        end_marker,
        start_name_map,
        end_name_map,
        dynamic_axes,
        save_file,
    )



# Copyright (c) OpenMMLab. All rights reserved.
# Modified from:
# https://github.com/pytorch/pytorch/blob/9ade03959392e5a90b74261012de1d806cab2253/torch/onnx/symbolic_opset9.py
from torch.onnx.symbolic_helper import parse_args

from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER
from juxtapose.mmdeploy.utils import Backend


@parse_args("v", "v", "f", "f", "i", "i")
def linear_no_bias(g, input, weight):
    """Symbolic function for `linear` without bias.

    PyTorch `nn.Linear` will be exported as ONNX node 'Gemm'.
    """
    return g.op("Gemm", input, weight, alpha_f=1.0, beta_f=1.0, transA_i=0, transB_i=1)


@parse_args("v", "v", "v", "f", "f", "i", "i")
def linear_normal(g, input, weight, bias):
    """Symbolic function for `linear`.

    PyTorch `nn.Linear` will be exported as ONNX node 'Gemm'.
    """
    return g.op(
        "Gemm", input, weight, bias, alpha_f=1.0, beta_f=1.0, transA_i=0, transB_i=1
    )


@SYMBOLIC_REWRITER.register_symbolic(
    "linear", is_pytorch=True, backend=Backend.NCNN.value
)
def linear__ncnn(g, input, weight, bias):
    """Support export linear This rewrite enable export Gemm."""
    if bias is None:
        return linear_no_bias(g, input, weight)
    else:
        return linear_normal(g, input, weight, bias)



# Copyright (c) OpenMMLab. All rights reserved.
MMDET_PARTITION_CFG = dict(
    single_stage=[
        dict(
            save_file='partition0.onnx',
            start='detector_forward:input',
            end='multiclass_nms:input',
            dynamic_axes={
                'input': {
                    0: 'batch',
                    2: 'height',
                    3: 'width'
                },
                'scores': {
                    0: 'batch',
                    1: 'num_boxes',
                },
                'boxes': {
                    0: 'batch',
                    1: 'num_boxes',
                },
            },
        )
    ],
    two_stage=[
        dict(
            save_file='partition0.onnx',
            start='detector_forward:input',
            end=['extract_feat:output', 'multiclass_nms[0]:input'],
            dynamic_axes={
                'input': {
                    0: 'batch',
                    2: 'height',
                    3: 'width'
                },
                'scores': {
                    0: 'batch',
                    1: 'num_boxes',
                },
                'boxes': {
                    0: 'batch',
                    1: 'num_boxes',
                },
            },
        ),
        dict(
            save_file='partition1.onnx',
            start='roi_extractor:output',
            end='bbox_head_forward:output',
            dynamic_axes={
                'bbox_feats': {
                    0: 'batch'
                },
                'cls_score': {
                    0: 'batch'
                },
                'bbox_pred': {
                    0: 'batch'
                },
            },
        )
    ])



# Copyright (c) OpenMMLab. All rights reserved.

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    "mmpose.models.heads.heatmap_heads.CPMHead.forward"
)
@FUNCTION_REWRITER.register_rewriter(
    "mmpose.models.heads.heatmap_heads.MSPNHead.forward"
)
def mspn_head__forward(self, feats):
    """Rewrite `forward` of MSPNHead and CPMHead for default backend.

    1. return last stage heatmaps directly.

    Args:
        feats (tuple[Tensor]): Input features.

    Returns:
        output_heatmap (torch.Tensor): Output heatmaps.
    """
    ctx = FUNCTION_REWRITER.get_context()
    msmu_batch_heatmaps = ctx.origin_func(self, feats)
    batch_heatmaps = msmu_batch_heatmaps[-1]
    return batch_heatmaps



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Tuple

import torch
from mmengine.config import ConfigDict
from torch import Tensor

from juxtapose.mmdeploy.codebase.mmdet import get_post_processing_params
from juxtapose.mmdeploy.core import FUNCTION_REWRITER
from juxtapose.mmdeploy.mmcv.ops.nms import multiclass_nms
from juxtapose.mmdeploy.utils import Backend, get_backend


@FUNCTION_REWRITER.register_rewriter(
    func_name="models.yolox_pose_head." "YOLOXPoseHead.predict"
)
def predict(self, x: Tuple[Tensor], batch_data_samples=None, rescale: bool = True):
    """Get predictions and transform to bbox and keypoints results.
    Args:
        x (Tuple[Tensor]): The input tensor from upstream network.
        batch_data_samples: Batch image meta info. Defaults to None.
        rescale: If True, return boxes in original image space.
            Defaults to False.

    Returns:
        Tuple[Tensor]: Predict bbox and keypoint results.
        - dets (Tensor): Predict bboxes and scores, which is a 3D Tensor,
            has shape (batch_size, num_instances, 5), the last dimension 5
            arrange as (x1, y1, x2, y2, score).
        - pred_kpts (Tensor): Predict keypoints and scores, which is a 4D
            Tensor, has shape (batch_size, num_instances, num_keypoints, 5),
            the last dimension 3 arrange as (x, y, score).
    """
    outs = self(x)
    predictions = self.predict_by_feat(
        *outs, batch_img_metas=batch_data_samples, rescale=rescale
    )
    return predictions


@FUNCTION_REWRITER.register_rewriter(
    func_name="models.yolox_pose_head." "YOLOXPoseHead.predict_by_feat"
)
def yolox_pose_head__predict_by_feat(
    self,
    cls_scores: List[Tensor],
    bbox_preds: List[Tensor],
    objectnesses: Optional[List[Tensor]] = None,
    kpt_preds: Optional[List[Tensor]] = None,
    vis_preds: Optional[List[Tensor]] = None,
    batch_img_metas: Optional[List[dict]] = None,
    cfg: Optional[ConfigDict] = None,
    rescale: bool = True,
    with_nms: bool = True,
) -> Tuple[Tensor]:
    """Transform a batch of output features extracted by the head into bbox and
    keypoint results.

    In addition to the base class method, keypoint predictions are also
    calculated in this method.

    Args:
        cls_scores (List[Tensor]): Classification scores for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * num_classes, H, W).
        bbox_preds (List[Tensor]): Box energies / deltas for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_priors * 4, H, W).
        objectnesses (Optional[List[Tensor]]): Score factor for
            all scale level, each is a 4D-tensor, has shape
            (batch_size, 1, H, W).
        kpt_preds (Optional[List[Tensor]]): Keypoints for all
            scale levels, each is a 4D-tensor, has shape
            (batch_size, num_keypoints * 2, H, W)
        vis_preds (Optional[List[Tensor]]): Keypoints scores for
            all scale levels, each is a 4D-tensor, has shape
            (batch_size, num_keypoints, H, W)
        batch_img_metas (Optional[List[dict]]): Batch image meta
            info. Defaults to None.
        cfg (Optional[ConfigDict]): Test / postprocessing
            configuration, if None, test_cfg would be used.
            Defaults to None.
        rescale (bool): If True, return boxes in original image space.
            Defaults to False.
        with_nms (bool): If True, do nms before return boxes.
            Defaults to True.
    Returns:
        Tuple[Tensor]: Predict bbox and keypoint results.
        - dets (Tensor): Predict bboxes and scores, which is a 3D Tensor,
            has shape (batch_size, num_instances, 5), the last dimension 5
            arrange as (x1, y1, x2, y2, score).
        - pred_kpts (Tensor): Predict keypoints and scores, which is a 4D
            Tensor, has shape (batch_size, num_instances, num_keypoints, 5),
            the last dimension 3 arrange as (x, y, score).
    """
    ctx = FUNCTION_REWRITER.get_context()
    deploy_cfg = ctx.cfg
    dtype = cls_scores[0].dtype
    device = cls_scores[0].device
    bbox_decoder = self.bbox_coder.decode

    assert len(cls_scores) == len(bbox_preds)
    cfg = self.test_cfg if cfg is None else cfg

    num_imgs = cls_scores[0].shape[0]
    featmap_sizes = [cls_score.shape[2:] for cls_score in cls_scores]

    self.mlvl_priors = self.prior_generator.grid_priors(
        featmap_sizes, dtype=dtype, device=device
    )

    flatten_priors = torch.cat(self.mlvl_priors)

    mlvl_strides = [
        flatten_priors.new_full(
            (featmap_size[0] * featmap_size[1] * self.num_base_priors,), stride
        )
        for featmap_size, stride in zip(featmap_sizes, self.featmap_strides)
    ]
    flatten_stride = torch.cat(mlvl_strides)

    # flatten cls_scores, bbox_preds and objectness
    flatten_cls_scores = [
        cls_score.permute(0, 2, 3, 1).reshape(num_imgs, -1, self.num_classes)
        for cls_score in cls_scores
    ]
    cls_scores = torch.cat(flatten_cls_scores, dim=1).sigmoid()

    flatten_bbox_preds = [
        bbox_pred.permute(0, 2, 3, 1).reshape(num_imgs, -1, 4)
        for bbox_pred in bbox_preds
    ]
    flatten_bbox_preds = torch.cat(flatten_bbox_preds, dim=1)

    if objectnesses is not None:
        flatten_objectness = [
            objectness.permute(0, 2, 3, 1).reshape(num_imgs, -1)
            for objectness in objectnesses
        ]
        flatten_objectness = torch.cat(flatten_objectness, dim=1).sigmoid()
        cls_scores = cls_scores * (flatten_objectness.unsqueeze(-1))

    scores = cls_scores
    bboxes = bbox_decoder(flatten_priors[None], flatten_bbox_preds, flatten_stride)

    # deal with key-poinsts
    priors = torch.cat(self.mlvl_priors)
    strides = [
        priors.new_full((featmap_size.numel() * self.num_base_priors,), stride)
        for featmap_size, stride in zip(featmap_sizes, self.featmap_strides)
    ]
    strides = torch.cat(strides)
    kpt_preds = torch.cat(
        [
            kpt_pred.permute(0, 2, 3, 1).reshape(num_imgs, -1, self.num_keypoints * 2)
            for kpt_pred in kpt_preds
        ],
        dim=1,
    )
    flatten_decoded_kpts = self.decode_pose(priors, kpt_preds, strides)

    vis_preds = torch.cat(
        [
            vis_pred.permute(0, 2, 3, 1).reshape(num_imgs, -1, self.num_keypoints, 1)
            for vis_pred in vis_preds
        ],
        dim=1,
    ).sigmoid()

    pred_kpts = torch.cat([flatten_decoded_kpts, vis_preds], dim=3)

    backend = get_backend(deploy_cfg)
    if backend == Backend.TENSORRT:
        # pad for batched_nms because its output index is filled with -1
        bboxes = torch.cat(
            [bboxes, bboxes.new_zeros((bboxes.shape[0], 1, bboxes.shape[2]))], dim=1
        )
        scores = torch.cat([scores, scores.new_zeros((scores.shape[0], 1, 1))], dim=1)
        pred_kpts = torch.cat(
            [
                pred_kpts,
                pred_kpts.new_zeros(
                    (pred_kpts.shape[0], 1, pred_kpts.shape[2], pred_kpts.shape[3])
                ),
            ],
            dim=1,
        )

    # nms
    post_params = get_post_processing_params(deploy_cfg)
    max_output_boxes_per_class = post_params.max_output_boxes_per_class
    iou_threshold = cfg.nms.get("iou_threshold", post_params.iou_threshold)
    score_threshold = cfg.get("score_thr", post_params.score_threshold)
    pre_top_k = post_params.get("pre_top_k", -1)
    keep_top_k = cfg.get("max_per_img", post_params.keep_top_k)
    # do nms
    _, _, nms_indices = multiclass_nms(
        bboxes,
        scores,
        max_output_boxes_per_class,
        iou_threshold,
        score_threshold,
        pre_top_k=pre_top_k,
        keep_top_k=keep_top_k,
        output_index=True,
    )

    batch_inds = torch.arange(num_imgs, device=scores.device).view(-1, 1)
    dets = torch.cat([bboxes, scores], dim=2)
    dets = dets[batch_inds, nms_indices, ...]
    pred_kpts = pred_kpts[batch_inds, nms_indices, ...]

    return dets, pred_kpts



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta

from mmengine import Config
from mmengine.registry import Registry

from juxtapose.mmdeploy.utils import Codebase, Task, get_task_type
from .task import BaseTask


class MMCodebase(metaclass=ABCMeta):
    """Wrap the apis of OpenMMLab Codebase."""

    task_registry: Registry = None

    def __init__(self) -> None:
        pass

    @classmethod
    def get_task_class(cls, task: Task) -> BaseTask:
        """Get the task processors class according to the task type.

        Args:
            task (Task): The task enumeration.

        Returns:
            type: The task processor class.
        """
        return cls.task_registry.module_dict[task.value]

    @classmethod
    def build_task_processor(cls, model_cfg: Config, deploy_cfg: Config, device: str):
        """The interface to build the task processors of the codebase.

        Args:
            model_cfg (str | Config): Model config file.
            deploy_cfg (str | Config): Deployment config file.
            device (str): A string specifying device type.

        Returns:
            BaseTask: A task processor.
        """
        task = get_task_type(deploy_cfg)
        return cls.task_registry.build(
            dict(
                type=task.value,
                model_cfg=model_cfg,
                deploy_cfg=deploy_cfg,
                device=device,
            )
        )

    @classmethod
    def register_deploy_modules(cls):
        """register deploy module."""
        raise NotImplementedError("register_deploy_modules not implemented.")

    @classmethod
    def register_all_modules(cls):
        """register codebase module."""
        raise NotImplementedError("register_all_modules not implemented.")


# Note that the build function returns the class instead of its instance.

CODEBASE = Registry("Codebases")


def get_codebase_class(codebase: Codebase) -> MMCodebase:
    """Get the codebase class from the registry.

    Args:
        codebase (Codebase): The codebase enum type.

    Returns:
        type: The codebase class
    """
    import importlib

    try:
        importlib.import_module(f"juxtapose.mmdeploy.codebase.{codebase.value}.deploy")
    except ImportError as e:
        from juxtapose.mmdeploy.utils import get_root_logger

        logger = get_root_logger()
        logger.warn(
            f"import juxtapose.mmdeploy.codebase.{codebase.value}.deploy failed"
            "Please check whether the module is the custom module."
            f"{e}"
        )
    return CODEBASE.build({"type": codebase.value})



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from functools import partial
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import torch

from juxtapose.mmdeploy.core import RewriterContext, patch_model
from juxtapose.mmdeploy.utils import IR, Backend, get_ir_config, get_root_logger
from ..core import PIPELINE_MANAGER


@PIPELINE_MANAGER.register_pipeline()
def trace(
    func: torch.nn.Module,
    inputs: Union[torch.Tensor, Tuple],
    output_path_prefix: Optional[str] = None,
    backend: Union[Backend, str] = "default",
    input_metas: Optional[Dict] = None,
    context_info: Dict = dict(),
    check_trace: bool = True,
    check_tolerance: float = 1e-05,
) -> torch.jit.TracedModule:
    """A wrapper of `torch.jit.trace` with some enhancement.

    Examples:
        >>> from juxtapose.mmdeploy.apis.torch_jit import trace
        >>>
        >>> func = create_model()
        >>> inputs = get_input_tensor()
        >>>
        >>> jit_model = trace(
        >>>     func,
        >>>     inputs,
        >>>     backend='torchscript',
        >>>     check_trace=False)
        >>>

    Args:
        func (torch.nn.Module): A Python function or `torch.nn.Module` that
            will be run with `example_inputs`.
        inputs (torch.Tensor, Tuple): A tuple of example inputs that will be
            passed to the function while tracing.
        output_path_prefix (str): The model would be serialized in
            `<output_path_prefix>.pth`, None if you don't want to
            save the model.
        backend (Backend|str): Which backend will the graph be used. Different
            backend would generate different graph.
        input_metas (Dict): The constant inputs of the model.
        context_info (Dict): The information that would be used in the context
            of exporting.
        check_trace (bool): Check if the same inputs run through traced code
            produce the same outputs.
        check_tolerance (float): Floating-point comparison tolerance to use in
            the checker procedure.

    Returns:
        torch.jit.TracedModule: The traced torch jit model.
    """
    logger = get_root_logger()
    logger.info("Export PyTorch model to torchscript.")

    def _add_or_update(cfg: dict, key: str, val: Any):
        if key in cfg and isinstance(cfg[key], dict) and isinstance(val, dict):
            cfg[key].update(val)
        else:
            cfg[key] = val

    context_info = deepcopy(context_info)
    deploy_cfg = context_info.pop("deploy_cfg", dict())
    ir_config = dict(type="torchscript")
    _add_or_update(deploy_cfg, "ir_config", ir_config)

    if isinstance(backend, Backend):
        backend = backend.value
    backend_config = dict(type=backend)
    _add_or_update(deploy_cfg, "backend_config", backend_config)

    context_info["cfg"] = deploy_cfg
    if "backend" not in context_info:
        context_info["backend"] = backend
    elif context_info["backend"] != backend:
        logger.warning(
            f'Find backend {context_info["backend"]} in context_info.'
            f" Expect {backend}."
        )
    if "ir" not in context_info:
        context_info["ir"] = IR.TORCHSCRIPT
    elif context_info["ir"] != backend:
        logger.warning(
            f'Find ir {context_info["ir"]} in context_info.'
            f" Expect {IR.TORCHSCRIPT}."
        )

    # patch model
    if isinstance(func, torch.nn.Module):
        ir = IR.get(get_ir_config(deploy_cfg)["type"])
        func = patch_model(func, cfg=deploy_cfg, backend=backend, ir=ir)

    with RewriterContext(**context_info), torch.no_grad():
        # patch input_metas
        if input_metas is not None:
            assert isinstance(
                input_metas, dict
            ), f"Expect input_metas type is dict, get {type(input_metas)}."
            model_forward = func.forward
            func.forward = partial(func.forward, **input_metas)

        # for exporting models with weight that depends on inputs
        func(*inputs) if isinstance(inputs, Sequence) else func(inputs)
        ts_model = torch.jit.trace(
            func, inputs, check_trace=check_trace, check_tolerance=check_tolerance
        )

        if input_metas is not None:
            func.forward = model_forward

    # save model
    if output_path_prefix is not None:
        output_path = output_path_prefix + ".pt"
        logger.info(f"Save PyTorch model: {output_path}.")
        torch.jit.save(ts_model, output_path)

    return ts_model



_base_ = 'mmdet::rtmdet/rtmdet_l_8xb32-300e_coco.py'
checkpoint = 'https://download.openmmlab.com/mmdetection/v3.0/rtmdet/cspnext_rsb_pretrain/cspnext-s_imagenet_600e.pth'  # noqa
model = dict(
    backbone=dict(
        deepen_factor=0.33,
        widen_factor=0.5,
        init_cfg=dict(
            type='Pretrained', prefix='backbone.', checkpoint=checkpoint)),
    neck=dict(in_channels=[128, 256, 512], out_channels=128, num_csp_blocks=1),
    bbox_head=dict(in_channels=128, feat_channels=128, exp_on_reg=False))

train_pipeline = [
    dict(type='LoadImageFromFile', backend_args={{_base_.backend_args}}),
    dict(type='LoadAnnotations', with_bbox=True),
    dict(type='CachedMosaic', img_scale=(640, 640), pad_val=114.0),
    dict(
        type='RandomResize',
        scale=(1280, 1280),
        ratio_range=(0.5, 2.0),
        keep_ratio=True),
    dict(type='RandomCrop', crop_size=(640, 640)),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    dict(type='Pad', size=(640, 640), pad_val=dict(img=(114, 114, 114))),
    dict(
        type='CachedMixUp',
        img_scale=(640, 640),
        ratio_range=(1.0, 1.0),
        max_cached_images=20,
        pad_val=(114, 114, 114)),
    dict(type='PackDetInputs')
]

train_pipeline_stage2 = [
    dict(type='LoadImageFromFile', backend_args={{_base_.backend_args}}),
    dict(type='LoadAnnotations', with_bbox=True),
    dict(
        type='RandomResize',
        scale=(640, 640),
        ratio_range=(0.5, 2.0),
        keep_ratio=True),
    dict(type='RandomCrop', crop_size=(640, 640)),
    dict(type='YOLOXHSVRandomAug'),
    dict(type='RandomFlip', prob=0.5),
    dict(type='Pad', size=(640, 640), pad_val=dict(img=(114, 114, 114))),
    dict(type='PackDetInputs')
]

train_dataloader = dict(dataset=dict(pipeline=train_pipeline))

custom_hooks = [
    dict(
        type='EMAHook',
        ema_type='ExpMomentumEMA',
        momentum=0.0002,
        update_buffers=True,
        priority=49),
    dict(
        type='PipelineSwitchHook',
        switch_epoch=280,
        switch_pipeline=train_pipeline_stage2)
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch


class NcnnDetectionOutputOp(torch.autograd.Function):
    """Create DetectionOutput op.

    A dummy DetectionOutput operator for ncnn end2end deployment.
    It will map to the DetectionOutput op of ncnn. After converting
    to ncnn, DetectionOutput op of ncnn will get called
    automatically.

    Args:
        loc (Tensor): The predicted boxes location tensor.
        conf (Tensor): The predicted boxes confidence of
            num_classes.
        anchor (Tensor): The prior anchors.
        score_threshold (float): Threshold of object
            score.
            Default: 0.35.
        nms_threshold (float): IoU threshold for NMS.
            Default: 0.45.
        nms_top_k (int): Number of bboxes after NMS.
            Default: 100.
        keep_top_k (int): Max number of bboxes of detection result.
            Default: 100.
        num_class (int): Number of classes, includes the background
            class.
            Default: 81.
    """

    @staticmethod
    def symbolic(g,
                 loc,
                 conf,
                 anchor,
                 score_threshold=0.35,
                 nms_threshold=0.45,
                 nms_top_k=100,
                 keep_top_k=100,
                 num_class=81,
                 target_stds=[0.1, 0.1, 0.2, 0.2]):
        """Symbolic function of dummy onnx DetectionOutput op for ncnn."""
        return g.op(
            'mmdeploy::DetectionOutput',
            loc,
            conf,
            anchor,
            score_threshold_f=score_threshold,
            nms_threshold_f=nms_threshold,
            nms_top_k_i=nms_top_k,
            keep_top_k_i=keep_top_k,
            num_class_i=num_class,
            vars_f=target_stds,
            outputs=1)

    @staticmethod
    def forward(ctx,
                loc,
                conf,
                anchor,
                score_threshold=0.35,
                nms_threshold=0.45,
                nms_top_k=100,
                keep_top_k=100,
                num_class=81,
                target_stds=[0.1, 0.1, 0.2, 0.2]):
        """Forward function of dummy onnx DetectionOutput op for ncnn."""
        return torch.rand(1, 100, 6)


ncnn_detection_output_forward = NcnnDetectionOutputOp.apply



# Copyright (c) OpenMMLab. All rights reserved.
import glob
import logging
import os
import sys
import traceback
from typing import Callable, Optional, Union

try:
    from torch import multiprocessing as mp
except ImportError:
    import multiprocess as mp

from juxtapose.mmdeploy.utils.logging import get_logger


def target_wrapper(
    target: Callable,
    log_level: int,
    ret_value: Optional[mp.Value] = None,
    *args,
    **kwargs,
):
    """The wrapper used to start a new subprocess.

    Args:
        target (Callable): The target function to be wrapped.
        log_level (int): Log level for logging.
        ret_value (mp.Value): The success flag of target.

    Return:
        Any: The return of target.
    """
    logger = logging.getLogger()
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d:%H:%M:%S",
    )
    logger.level
    logger.setLevel(log_level)
    if ret_value is not None:
        ret_value.value = -1
    try:
        result = target(*args, **kwargs)
        if ret_value is not None:
            ret_value.value = 0
        return result
    except Exception as e:
        logging.error(e)
        traceback.print_exc(file=sys.stdout)


def get_root_logger(log_file=None, log_level=logging.INFO) -> logging.Logger:
    """Get root logger.

    Args:
        log_file (str, optional): File path of log. Defaults to None.
        log_level (int, optional): The level of logger.
            Defaults to logging.INFO.
    Returns:
        logging.Logger: The obtained logger
    """
    logger = get_logger(name="mmdeploy", log_file=log_file, log_level=log_level)

    return logger


def deprecate(
    status: str = "future",
    dst_obj: Optional[Union[object, str]] = None,
    msg: str = "",
    *args,
    **kwargs,
) -> None:
    """Deprecate a function or a class.

    Args:
        status (str, optional): The status of the function or class.
            Defaults to future.
        dst_obj (str, object, optional): The object that will replace
            the original one. Defaults to None.
        msg (str): Additional message to be printed.

    Examples:
        >>> from math import ceil
        >>> from juxtapose.mmdeploy.utils.utils import deprecate
        >>> @deprecate(status='past', dst_obj=ceil, msg='')
        >>> def my_ceil(num):
        >>>     num = num if(num==int(num)) else int(num) + 1
        >>>     return num
    """
    logger = get_root_logger()

    def _register(src_obj):
        def fun(*args, **kwargs):
            if status == "future":
                logger.warning(
                    f"DeprecationWarning: {src_obj.__name__} will be "
                    f"deprecated in the future. {msg}"
                )
            elif status == "past":
                assert dst_obj is not None, "for deprecated object, there"
                " must be a destination object"
                logger.warning(
                    f"DeprecationWarning: {src_obj.__name__} was deprecated,"
                    f" use {dst_obj.__name__} instead. {msg}"
                )
            else:
                raise KeyError(f"Unexpected key {status}")
            result = src_obj(*args, **kwargs)
            return result

        return fun

    return _register


def get_file_path(prefix, candidates) -> str:
    """Search for file in candidates.

    Args:
        prefix (str): Prefix of the paths.
        candidates (str): Candidate paths
    Returns:
        str: file path or '' if not found
    """
    for candidate in candidates:
        wildcard = os.path.abspath(os.path.join(prefix, candidate))
        paths = glob.glob(wildcard)
        if paths:
            lib_path = paths[0]
            return lib_path
    return ""



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence

import torch
from packaging.version import parse

from juxtapose.mmdeploy.core import FUNCTION_REWRITER, SYMBOLIC_REWRITER


@FUNCTION_REWRITER.register_rewriter(func_name="torch.Tensor.__setitem__")
def tensor__setitem__default(self, key, value):
    """Rewrite `setitem` to ease the index put."""
    ctx = FUNCTION_REWRITER.get_context()

    # only support torch>=1.9.0
    if parse(torch.__version__) < parse("1.9.0"):
        return ctx.origin_func(self, key, value)

    if isinstance(key, slice):
        key = (key,)

    if not isinstance(key, Sequence):
        return ctx.origin_func(self, key, value)

    for k in key:
        if not isinstance(k, slice) or k.step is not None:
            return ctx.origin_func(self, key, value)

    out = value

    # value could be scalar or single value Tensor
    self_shape = self.shape
    out_shape = list(self_shape)
    for i, k in enumerate(key):
        start = 0 if k.start is None else k.start
        start = start if start >= 0 else self_shape[i] + start
        stop = self_shape[i] if k.stop is None else k.stop
        stop = stop if stop >= 0 else self_shape[i] + stop
        out_shape[i] = stop - start

    if not isinstance(out, torch.Tensor):
        out = self.new_full(out_shape, out)
    elif out.numel() == 1:
        out = out.expand(out_shape)

    for i, k in enumerate(key):
        if k == slice(None):
            continue

        cat_list = []

        # slice self start
        if k.start is not None:
            self_slice_start = (slice(None),) * i + (slice(0, k.start),) + key[i + 1 :]
            self_start = self[self_slice_start]
            cat_list.append(self_start)

        # add value
        cat_list.append(out)

        # slice self end
        if k.stop is not None:
            self_slice_end = (slice(None),) * i + (slice(k.stop, None),) + key[i + 1 :]
            self_end = self[self_slice_end]
            cat_list.append(self_end)

        # concate
        out = torch.cat(cat_list, dim=i)

    # self assign
    # Note that set item does not return any value
    self[...] = out


if parse(torch.__version__) >= parse("1.12.0"):

    @SYMBOLIC_REWRITER.register_symbolic("copy", is_pytorch=True)
    def copy__default(g, x, y, non_blocking):
        return x



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Iterable

import torch

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    func_name="torch.Tensor.__getitem__", backend="ascend"
)
def tensor__getitem__ascend(self, key) -> torch.Tensor:
    """Rewrite `getitem` for ascend backend.

    Ascend does not support negative select
    """
    ctx = FUNCTION_REWRITER.get_context()
    if not isinstance(key, (tuple, list)):
        if isinstance(key, int) and key < 0:
            key = self.dim() + key
        return ctx.origin_func(self, key)

    def _num_slice_types(slices):
        num_slice = 0
        for s in slices:
            if isinstance(s, slice) or isinstance(s, int) or isinstance(s, Iterable):
                num_slice += 1
        return num_slice

    shape = self.shape
    new_key = list(key)
    num_ellipsis = len(shape) - _num_slice_types(new_key)
    dim_count = 0
    for i, k in enumerate(new_key):
        if isinstance(k, int):
            if k < 0:
                new_key[i] = shape[dim_count] + k
        if k == Ellipsis:
            dim_count = dim_count + num_ellipsis
        elif k is not None:
            dim_count += 1
    return ctx.origin_func(self, new_key)



# Copyright (c) OpenMMLab. All rights reserved.
import os

from juxtapose.mmdeploy.utils import get_file_path


def get_ops_path() -> str:
    """Get the library path of onnxruntime custom ops.

    Returns:
        str: The library path to onnxruntime custom ops.
    """
    candidates = [
        "../../lib/libmmdeploy_onnxruntime_ops.so",
        "../../lib/mmdeploy_onnxruntime_ops.dll",
    ]
    return get_file_path(os.path.dirname(__file__), candidates)



# Copyright (c) OpenMMLab. All rights reserved.
import torch


class NcnnPriorBoxOp(torch.autograd.Function):
    """Create PriorBox op for ncnn.

    A dummy PriorBox operator for ncnn end2end deployment. It
    will map to the PriorBox op of ncnn. After converting to ncnn,
    PriorBox op of ncnn will get called automatically.

    Args:
        feat (Tensor): Feature maps to generate prior boxes.
        aspect_ratios (List[float]): The list of ratios between
            the height and width of anchors in a single level.
            Default: [2., 3.].
        image_height: (int): The height of the input image.
            Default: 320.
        image_width: (int): The width of the input image.
            Default: 320.
        max_sizes: (List[double]): The list of maximum anchor
            sizes on each level.
            Default: [320.]
        min_sizes: (List[double]): The list of minimum anchor
            sizes on each level.
            Default: [304.]
        offset: (float): The offset of center in proportion to
            anchors' width and height. It is not the same as the
            'center_offset' in
            mmdet.core.anchor.anchor_generator.py.
            Default: 0.5.
        step_mmdetection: (int): The boolean variable of whether
            to reproduce the strides parameters of mmdetection
            in ncnn PriorBox layer implementation.
            Default: 1.
    """

    @staticmethod
    def symbolic(g,
                 feat,
                 aspect_ratios=[2, 3],
                 image_height=300,
                 image_width=300,
                 step_height=300,
                 step_width=300,
                 max_sizes=[300],
                 min_sizes=[285],
                 offset=0.5,
                 step_mmdetection=1):
        """Symbolic function of dummy onnx PriorBox op for ncnn."""
        return g.op(
            'mmdeploy::PriorBox',
            feat,
            aspect_ratios_f=aspect_ratios,
            image_height_i=image_height,
            image_width_i=image_width,
            step_height_f=step_height,
            step_width_f=step_width,
            max_sizes_f=max_sizes,
            min_sizes_f=min_sizes,
            offset_f=offset,
            step_mmdetection_i=step_mmdetection,
            outputs=1)

    @staticmethod
    def forward(ctx,
                feat,
                aspect_ratios=[2, 3],
                image_height=300,
                image_width=300,
                step_height=300,
                step_width=300,
                max_sizes=[300],
                min_sizes=[285],
                offset=0.5,
                step_mmdetection=1):
        """Forward function of dummy onnx PriorBox op for ncnn."""
        num_priors = len(aspect_ratios) * 2 + len(min_sizes) + \
            len(max_sizes)
        return torch.rand(2, 4 * num_priors * feat.shape[-1] * feat.shape[-2])


ncnn_prior_box_forward = NcnnPriorBoxOp.apply



# Copyright (c) OpenMMLab. All rights reserved.

import copy
import os.path as osp
import random
import string
import tempfile
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pytest
import torch
from mmengine import Config
from mmengine.model import BaseModel
from torch import nn

try:
    from torch.testing import assert_close as torch_assert_close
except Exception:
    from torch.testing import assert_allclose as torch_assert_close

import juxtapose.mmdeploy.codebase  # noqa: F401,F403
from juxtapose.mmdeploy.core import RewriterContext, patch_model
from juxtapose.mmdeploy.utils import (
    IR,
    Backend,
    get_backend,
    get_dynamic_axes,
    get_ir_config,
    get_onnx_config,
)


def backend_checker(backend: Backend, require_plugin: bool = False):
    """A decorator which checks if a backend is available.

    Args:
        backend (Backend): The backend needs to be checked.
        require_plugin (bool): The checker will check if the backend package
            has been installed. If this variable is `True`, then the checker
            will also check if the backend plugin has been compiled. Default
            to `False`.
    """
    from juxtapose.mmdeploy.backend.base import get_backend_manager

    backend_mgr = get_backend_manager(backend.value)
    result = backend_mgr.is_available(with_custom_ops=require_plugin)

    checker = pytest.mark.skipif(
        not result, reason=f"{backend.value} package is not available"
    )

    return checker


def check_backend(backend: Backend, require_plugin: bool = False):
    """A utility to check if a backend is available.

    Args:
        backend (Backend): The backend needs to be checked.
        require_plugin (bool): The checker will check if the backend package
            has been installed. If this variable is `True`, then the checker
            will also check if the backend plugin has been compiled. Default
            to `False`.
    """
    from juxtapose.mmdeploy.backend.base import get_backend_manager

    backend_mgr = get_backend_manager(backend.value)
    result = backend_mgr.is_available(with_custom_ops=require_plugin)

    if not result:
        pytest.skip(f"{backend.value} package is not available")


class WrapFunction(nn.Module):
    """Wrap a pytorch function to nn.Module.

    It serves as partial function and can be exportable to ONNX.

    Args:
        wrapped_function (Callable): Input function to be wrapped.

    Examples:
        >>> from juxtapose.mmdeploy.utils.test import WrapFunction
        >>> import torch
        >>>
        >>> def clip(x, min, max):
        >>>     return torch.clamp(x, min, max)
        >>>
        >>> wrapped_model = WrapFunction(clip)
    """

    def __init__(self, wrapped_function: Callable, **kwargs):
        super(WrapFunction, self).__init__()
        self.wrapped_function = wrapped_function
        self.kwargs = kwargs

    def forward(self, *args, **kwargs) -> Any:
        """Call the wrapped function."""
        kwargs.update(self.kwargs)
        return self.wrapped_function(*args, **kwargs)


class WrapModel(nn.Module):
    """A wrapper class for rewrite unittests.

    It serves as partial function but can be rewritten with RewriterContext.

    Args:
        model (nn.Module): A pytorch module.
        func_name (str): Which function to use as forward function.

    Examples:
        >>> from juxtapose.mmdeploy.utils.test import WrapModel
        >>> from mmdet.models import AnchorHead
        >>>
        >>> model = AnchorHead(num_classes=4, in_channels=1)
        >>> wrapped_model = WrapModel(anchor_head,
        >>>                           'get_bboxes',
        >>>                           with_nms=True)
    """

    def __init__(self, model: nn.Module, func_name: str, **kwargs):
        super(WrapModel, self).__init__()
        assert hasattr(model, func_name), f"Got unexpected func name: {func_name}"
        self.model = model
        self.kwargs = kwargs
        self.func_name = func_name

    def forward(self, *args, **kwargs):
        """Run forward of the model."""
        kwargs.update(self.kwargs)
        func = getattr(self.model, self.func_name)
        return func(*args, **kwargs)


class DummyModel(BaseModel):
    """A dummy model for unit tests.

    Args:
        outputs (Any): Predefined output variables.
    """

    def __init__(self, outputs=None, *args, **kwargs):
        super().__init__()
        self.outputs = outputs

    def forward(self, *args, **kwargs):
        """Run forward."""
        return self.outputs


class SwitchBackendWrapper:
    """A switcher for backend wrapper for unit tests.
    Examples:
        >>> from juxtapose.mmdeploy.utils.test import SwitchBackendWrapper
        >>> from juxtapose.mmdeploy.backend.onnxruntime import ORTWrapper
        >>> with SwitchBackendWrapper(ORTWrapper) as wrapper:
        >>>     wrapper.set(ORTWrapper, outputs=outputs)
        >>>     ...
        >>>     # ORTWrapper will recover when exiting context
        >>> ...
    """

    init = None
    forward = None
    call = None

    class BackendWrapper(torch.nn.Module):
        """A dummy backend wrapper for unit tests.

        To enable BaseWrapper.output_to_list(), the wrapper needs member
        variable `_output_names` that is set in constructor. Therefore,
        the dummy BackendWrapper needs a constructor that receives
        output_names.

        Args:
            output_names (Any): `output_name` of BaseWrapper
        """

        def __init__(self, output_names=["dets", "labels"], *args, **kwargs):
            torch.nn.Module.__init__(self)
            self._output_names = output_names

        def forward(self, *args, **kwargs):
            """Run forward."""
            return self.outputs

        def __call__(self, *args, **kwds):
            """Call the forward method."""
            return self.forward(*args, **kwds)

    def __init__(self, recover_class):
        self._recover_class = recover_class

    def __enter__(self):
        return self

    def __exit__(self, type, value, trace):
        self.recover()

    def set(self, **kwargs):
        """Replace attributes in backend wrappers with dummy items."""
        obj = self._recover_class
        self.init = obj.__init__
        self.forward = obj.forward
        self.call = obj.__call__
        obj.__init__ = SwitchBackendWrapper.BackendWrapper.__init__
        obj.forward = SwitchBackendWrapper.BackendWrapper.forward
        obj.__call__ = SwitchBackendWrapper.BackendWrapper.__call__
        for k, v in kwargs.items():
            setattr(obj, k, v)

    def recover(self):
        """Recover to original class."""
        assert (
            self.init is not None and self.forward is not None
        ), "recover method must be called after exchange"
        obj = self._recover_class
        obj.__init__ = self.init
        obj.forward = self.forward
        obj.__call__ = self.call


def assert_allclose(
    expected: List[Union[torch.Tensor, np.ndarray]],
    actual: List[Union[torch.Tensor, np.ndarray]],
    tolerate_small_mismatch: bool = False,
):
    """Determine whether all actual values are closed with the expected values.

    Args:
        actual (list[torch.Tensor | np.ndarray): Actual value.
        expected (list[torch.Tensor | np.ndarray): Expected value.
        tolerate_small_mismatch (bool): Whether tolerate small mismatch,
        Default is False.
    """
    if not (isinstance(expected, list) and isinstance(actual, list)):
        raise ValueError("Argument desired and actual should be a list")
    if len(expected) != len(actual):
        raise ValueError("Length of desired and actual should be equal")

    for i in range(0, len(expected)):
        if isinstance(expected[i], (list, np.ndarray)):
            expected[i] = torch.tensor(expected[i])
        if isinstance(actual[i], (list, np.ndarray)):
            actual[i] = torch.tensor(actual[i])
        try:
            torch_assert_close(actual[i], expected[i], rtol=1e-03, atol=1e-05)
        except AssertionError as error:
            if tolerate_small_mismatch:
                assert "(0.00%)" in str(error), str(error)
            else:
                raise


def get_model_outputs(model: nn.Module, func_name: str, model_inputs: dict) -> Any:
    """To get outputs of pytorch model.

    Args:
        model (nn.Module): The input model.
        func_name (str): The function of model.
        model_inputs (dict): The inputs for model.

    Returns:
        Any: The output of model, decided by the input model.
    """
    assert hasattr(model, func_name), f"Got unexpected func name: {func_name}"
    func = getattr(model, func_name)
    model_outputs = func(**copy.deepcopy(model_inputs))
    return model_outputs


def get_flatten_inputs(
    model_inputs: Dict[str, Union[Tuple, List, torch.Tensor]]
) -> Dict[str, torch.Tensor]:
    """This function unwraps lists and tuples from 'model_inputs' and assigns a
    unique name to their values.

    Example:
        model_inputs = {'A': list(tensor_0, tensor_1), 'B': tensor_3}
        flatten_inputs = get_flatten_inputs(model_inputs)
        flatten_inputs: {'A_0': tensor_0, 'A_1': tensor_1, 'B': tensor_3}

    Args:
        model_inputs (dict): Key-value pairs of model inputs with
            lists and tuples.

    Returns:
        Dict[str, torch.Tensor]: Key-value pairs of model inputs with
            unwrapped lists and tuples.
    """
    flatten_inputs = {}
    for name, value in model_inputs.items():
        if isinstance(value, torch.Tensor):
            flatten_inputs[name] = value
        elif isinstance(value, (list, tuple)):
            if len(value) == 1:
                flatten_inputs[name] = value[0]
            else:
                for i, tensor in enumerate(value):
                    name_i = f"{name}_{i}"
                    flatten_inputs[name_i] = tensor
    return flatten_inputs


def get_onnx_model(
    wrapped_model: nn.Module,
    model_inputs: Dict[str, Union[Tuple, List, torch.Tensor]],
    deploy_cfg: Config,
) -> str:
    """To get path to onnx model after export.

    Args:
        wrapped_model (nn.Module): The input model.
        model_inputs (dict): Inputs for model.
        deploy_cfg (Config): Deployment config.

    Returns:
        str: The path to the ONNX model file.
    """
    onnx_file_path = tempfile.NamedTemporaryFile(suffix=".onnx").name
    onnx_cfg = get_onnx_config(deploy_cfg)
    backend = get_backend(deploy_cfg)
    patched_model = patch_model(wrapped_model, cfg=deploy_cfg, backend=backend.value)
    flatten_model_inputs = get_flatten_inputs(model_inputs)
    input_names = onnx_cfg.get("input_names", None)
    if input_names is None:
        input_names = [k for k, v in flatten_model_inputs.items() if k != "ctx"]
    output_names = onnx_cfg.get("output_names", None)
    dynamic_axes = get_dynamic_axes(deploy_cfg, input_names)

    class DummyModel(torch.nn.Module):
        def __init__(self):
            super(DummyModel, self).__init__()
            self.model = patched_model

        def forward(self, inputs: dict):
            return self.model(**inputs)

    model = DummyModel().eval()

    with RewriterContext(
        cfg=deploy_cfg, backend=backend.value, opset=11
    ), torch.no_grad():
        torch.onnx.export(
            model,
            (model_inputs, {}),
            onnx_file_path,
            export_params=True,
            input_names=input_names,
            output_names=output_names,
            opset_version=11,
            dynamic_axes=dynamic_axes,
            keep_initializers_as_inputs=False,
        )
    return onnx_file_path


def get_ts_model(
    wrapped_model: nn.Module,
    model_inputs: Dict[str, Union[Tuple, List, torch.Tensor]],
    deploy_cfg: Config,
) -> str:
    """To get path to onnx model after export.

    Args:
        wrapped_model (nn.Module): The input model.
        model_inputs (dict): Inputs for model.
        deploy_cfg (Config): Deployment config.

    Returns:
        str: The path to the TorchScript model file.
    """
    ir_file_path = tempfile.NamedTemporaryFile(suffix=".pt").name
    backend = get_backend(deploy_cfg)

    from juxtapose.mmdeploy.apis.torch_jit import trace

    context_info = dict(deploy_cfg=deploy_cfg)
    output_prefix = osp.splitext(ir_file_path)[0]

    example_inputs = [v for _, v in model_inputs.items()]
    trace(
        wrapped_model,
        example_inputs,
        output_path_prefix=output_prefix,
        backend=backend,
        context_info=context_info,
    )
    return ir_file_path


def get_backend_outputs(
    ir_file_path: str,
    model_inputs: Dict[str, Union[Tuple, List, torch.Tensor]],
    deploy_cfg: Config,
) -> Union[Any, None]:
    """To get backend outputs of model.

    Args:
        ir_file_path (str): The path to the IR file.
        model_inputs (dict): Inputs for model.
        deploy_cfg (Config): Deployment config.

    Returns:
        Union[Any, None]: The outputs of model, decided by the backend wrapper.
            If the backend specified in 'deploy_cfg' is not available,
            then None will be returned.
    """
    from juxtapose.mmdeploy.apis.utils import to_backend

    backend = get_backend(deploy_cfg)
    flatten_model_inputs = get_flatten_inputs(model_inputs)
    ir_config = get_ir_config(deploy_cfg)
    input_names = ir_config.get("input_names", None)
    output_names = ir_config.get("output_names", None)
    if input_names is None:
        input_names = [k for k, v in flatten_model_inputs.items() if k != "ctx"]

    work_dir = tempfile.TemporaryDirectory().name
    device = "cpu"

    # TODO: Try to wrap these code into backend manager
    if backend != Backend.TORCHSCRIPT:
        model_inputs = flatten_model_inputs
    if backend == Backend.TENSORRT:
        device = "cuda"
        model_inputs = dict((k, v.cuda()) for k, v in model_inputs.items())
    elif backend == Backend.OPENVINO:
        input_info = {name: value.shape for name, value in flatten_model_inputs.items()}
        deploy_cfg["backend_config"]["model_inputs"] = [dict(opt_shapes=input_info)]
    backend_files = to_backend(
        backend.value,
        [ir_file_path],
        work_dir=work_dir,
        deploy_cfg=deploy_cfg,
        device=device,
    )
    backend_feats = model_inputs

    if backend == Backend.TORCHSCRIPT:
        backend_feats = [v for _, v in model_inputs.items()]

    from juxtapose.mmdeploy.codebase.base import BaseBackendModel

    backend_model = BaseBackendModel._build_wrapper(
        backend,
        backend_files,
        device,
        input_names=input_names,
        output_names=output_names,
    )
    with torch.no_grad():
        backend_outputs = backend_model(backend_feats)
    backend_outputs = backend_model.output_to_list(backend_outputs)
    return backend_outputs


def get_rewrite_outputs(
    wrapped_model: nn.Module,
    model_inputs: Dict[str, Union[Tuple, List, torch.Tensor]],
    deploy_cfg: Config,
    run_with_backend: bool = True,
) -> Tuple[Any, bool]:
    """To get outputs of generated onnx model after rewrite.

    Args:
        wrapped_model (nn.Module): The input model.
        model_inputs (dict): Inputs for model.
        deploy_cfg (Config): Deployment config.
        run_with_backend (bool): Whether to run inference with backend.
            Default is True.

    Returns:
        List[torch.Tensor]: The outputs of model.
        bool: A flag indicate the type of outputs. If the flag is True, then
        the outputs are backend output, otherwise they are outputs of wrapped
        pytorch model.
    """
    backend = get_backend(deploy_cfg)
    with RewriterContext(
        cfg=deploy_cfg, backend=backend.value, opset=11
    ), torch.no_grad():
        ctx_outputs = wrapped_model(**copy.deepcopy(model_inputs))

    ir_type = get_ir_config(deploy_cfg).get("type", None)
    if ir_type == IR.TORCHSCRIPT.value:
        ir_file_path = get_ts_model(wrapped_model, model_inputs, deploy_cfg)
    else:  # TODO onnx as default, make it strict when more IR types involved
        ir_file_path = get_onnx_model(wrapped_model, model_inputs, deploy_cfg)

    backend_outputs = None
    if run_with_backend:
        backend_outputs = get_backend_outputs(ir_file_path, model_inputs, deploy_cfg)

    if backend_outputs is None:
        return ctx_outputs, False
    else:
        return backend_outputs, True


def get_random_name(
    length: int = 10, seed: Optional[Union[int, float, str, bytes, bytearray]] = None
) -> str:
    """Generates a random string of the specified length. Can be used to
    generate random names for model inputs and outputs.

    Args:
        length (int): The number of characters in the string..
        seed (Optional[Union[int, float, str, bytes, bytearray]]):
            Seed for a random number generator.

    Returns:
        str: A randomly generated string that can be used as a name.
    """
    if seed:
        random_state = random.getstate()
        random.seed(seed)
    random_name = "".join(
        random.choices(string.ascii_letters + string.digits, k=length)
    )
    if seed:
        random.setstate(random_state)
    return random_name



# Copyright (c) OpenMMLab. All rights reserved.
from juxtapose.mmdeploy.core import SYMBOLIC_REWRITER


@SYMBOLIC_REWRITER.register_symbolic(
    "mmcv.ops.multi_scale_deform_attn.MultiScaleDeformableAttnFunction"
)
def ms_deform_attn_default(
    g,
    value,
    value_spatial_shapes,
    value_level_start_index,
    sampling_locations,
    attention_weights,
    im2col_step=64,
):
    """Rewrite msda symbolic function for all backend."""
    return g.op(
        "mmdeploy::MMCVMultiScaleDeformableAttention",
        value,
        value_spatial_shapes,
        value_level_start_index,
        sampling_locations,
        attention_weights,
        im2col_step_i=im2col_step,
    )



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta
from typing import Optional, Sequence, Union

import mmengine
from mmengine.model import BaseModel
from torch import nn

from juxtapose.mmdeploy.utils import Backend, get_ir_config


class BaseBackendModel(BaseModel, metaclass=ABCMeta):
    """A backend model wraps the details to initialize and run a backend
    engine."""

    def __init__(
        self,
        deploy_cfg: Optional[Union[str, mmengine.Config]] = None,
        data_preprocessor: Optional[Union[dict, nn.Module]] = None,
        *args,
        **kwargs,
    ):
        """The default for building the base class.

        Args:
            deploy_cfg (str | mmengine.Config | None): The deploy config.
        """
        input_names = output_names = None
        if deploy_cfg is not None:
            ir_config = get_ir_config(deploy_cfg)
            output_names = ir_config.get("output_names", None)
            input_names = ir_config.get("input_names", None)
        # TODO use input_names instead in the future for multiple inputs
        self.input_name = input_names[0] if input_names else "input"
        self.output_names = output_names if output_names else ["output"]
        super(BaseBackendModel, self).__init__(data_preprocessor=data_preprocessor)

    @staticmethod
    def _build_wrapper(
        backend: Backend,
        backend_files: Sequence[str],
        device: str,
        input_names: Optional[Sequence[str]] = None,
        output_names: Optional[Sequence[str]] = None,
        deploy_cfg: Optional[mmengine.Config] = None,
        *args,
        **kwargs,
    ):
        """The default methods to build backend wrappers.

        Args:
            backend (Backend): The backend enum type.
            backend_files (Sequence[str]): Paths to all required backend files(
                e.g. '.onnx' for ONNX Runtime, '.param' and '.bin' for ncnn).
            device (str): A string specifying device type.
            input_names (Sequence[str] | None): Names of model inputs in
                order. Defaults to `None`.
            output_names (Sequence[str] | None): Names of model outputs in
                order. Defaults to `None` and the wrapper will load the output
                names from the model.
            deploy_cfg: Deployment config file.
        """
        from juxtapose.mmdeploy.backend.base import get_backend_manager

        backend_mgr = get_backend_manager(backend.value)
        if backend_mgr is None:
            raise NotImplementedError(f"Unsupported backend type: {backend.value}")
        return backend_mgr.build_wrapper(
            backend_files, device, input_names, output_names, deploy_cfg, **kwargs
        )

    def destroy(self):
        if hasattr(self, "wrapper") and hasattr(self.wrapper, "destroy"):
            self.wrapper.destroy()



# Copyright (c) OpenMMLab. All rights reserved.
from .rewriter_manager import (FUNCTION_REWRITER, MODULE_REWRITER,
                               SYMBOLIC_REWRITER, RewriterContext, patch_model)

__all__ = [
    'FUNCTION_REWRITER',
    'RewriterContext',
    'MODULE_REWRITER',
    'patch_model',
    'SYMBOLIC_REWRITER',
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List

import torch
from torch import Tensor
from torch.nn import functional as F

from juxtapose.mmdeploy.core import FUNCTION_REWRITER


@FUNCTION_REWRITER.register_rewriter(
    "mmdet.models.dense_heads.DETRHead.predict_by_feat"
)
def detrhead__predict_by_feat__default(
    self,
    all_cls_scores_list: List[Tensor],
    all_bbox_preds_list: List[Tensor],
    batch_img_metas: List[dict],
    rescale: bool = True,
):
    """Rewrite `predict_by_feat` of `FoveaHead` for default backend."""
    from mmdet.structures.bbox import bbox_cxcywh_to_xyxy

    cls_scores = all_cls_scores_list[-1]
    bbox_preds = all_bbox_preds_list[-1]

    img_shape = batch_img_metas[0]["img_shape"]
    max_per_img = self.test_cfg.get("max_per_img", len(cls_scores[0]))
    batch_size = cls_scores.size(0)
    # `batch_index_offset` is used for the gather of concatenated tensor

    # supports dynamical batch inference
    if self.loss_cls.use_sigmoid:
        batch_index_offset = (
            torch.arange(batch_size).to(cls_scores.device) * max_per_img
        )
        batch_index_offset = batch_index_offset.unsqueeze(1).expand(
            batch_size, max_per_img
        )
        cls_scores = cls_scores.sigmoid()
        scores, indexes = cls_scores.flatten(1).topk(max_per_img, dim=1)
        det_labels = indexes % self.num_classes
        bbox_index = indexes // self.num_classes
        bbox_index = (bbox_index + batch_index_offset).view(-1)
        bbox_preds = bbox_preds.view(-1, 4)[bbox_index]
        bbox_preds = bbox_preds.view(batch_size, -1, 4)
    else:
        scores, det_labels = F.softmax(cls_scores, dim=-1)[..., :-1].max(-1)
        scores, bbox_index = scores.topk(max_per_img, dim=1)
        batch_inds = torch.arange(batch_size, device=scores.device).unsqueeze(-1)
        bbox_preds = bbox_preds[batch_inds, bbox_index, ...]
        # add unsqueeze to support tensorrt
        det_labels = det_labels.unsqueeze(-1)[batch_inds, bbox_index, ...].squeeze(-1)

    det_bboxes = bbox_cxcywh_to_xyxy(bbox_preds)

    if isinstance(img_shape, torch.Tensor):
        hw = img_shape.flip(0).to(det_bboxes.device)
    else:
        hw = det_bboxes.new_tensor([img_shape[1], img_shape[0]])
    shape_scale = torch.cat([hw, hw])
    shape_scale = shape_scale.view(1, 1, -1)
    det_bboxes = det_bboxes * shape_scale
    # dynamically clip bboxes
    x1, y1, x2, y2 = det_bboxes.split((1, 1, 1, 1), dim=-1)
    from juxtapose.mmdeploy.codebase.mmdet.deploy import clip_bboxes

    x1, y1, x2, y2 = clip_bboxes(x1, y1, x2, y2, img_shape)
    det_bboxes = torch.cat([x1, y1, x2, y2], dim=-1)
    det_bboxes = torch.cat((det_bboxes, scores.unsqueeze(-1)), -1)

    return det_bboxes, det_labels



# Copyright (c) OpenMMLab. All rights reserved.
import importlib

from .constants import IR, SDK_TASK_MAP, Backend, Codebase, Task
from .device import parse_cuda_device_id, parse_device_id, parse_device_type
from .env import get_backend_version, get_codebase_version, get_library_version
from .utils import get_file_path, get_root_logger, target_wrapper

__all__ = [
    'SDK_TASK_MAP', 'IR', 'Backend', 'Codebase', 'Task',
    'parse_cuda_device_id', 'get_library_version', 'get_codebase_version',
    'get_backend_version', 'parse_device_id', 'get_file_path',
    'get_root_logger', 'target_wrapper', 'parse_device_type'
]

if importlib.util.find_spec('mmcv') is not None:
    # yapf: disable
    from .config_utils import (cfg_apply_marks, get_backend,
                               get_backend_config, get_calib_config,
                               get_calib_filename, get_codebase,
                               get_codebase_config, get_common_config,
                               get_dynamic_axes, get_input_shape,
                               get_ir_config, get_model_inputs,
                               get_normalization, get_onnx_config,
                               get_partition_config, get_precision,
                               get_quantization_config, get_rknn_quantization,
                               get_task_type, is_dynamic_batch,
                               is_dynamic_shape, load_config)

    # yapf: enable

    __all__ += [
        'cfg_apply_marks', 'get_backend', 'get_backend_config',
        'get_calib_config', 'get_calib_filename', 'get_codebase',
        'get_codebase_config', 'get_common_config', 'get_dynamic_axes',
        'get_input_shape', 'get_ir_config', 'get_model_inputs',
        'get_onnx_config', 'get_partition_config', 'get_quantization_config',
        'get_precision', 'get_task_type', 'is_dynamic_batch',
        'is_dynamic_shape', 'load_config', 'get_rknn_quantization',
        'get_normalization'
    ]



# Copyright (c) OpenMMLab. All rights reserved.
from copy import deepcopy
from typing import Callable, Dict, Optional

import torch
from torch.utils.data import DataLoader

from ..core import PIPELINE_MANAGER


@PIPELINE_MANAGER.register_pipeline()
def create_calib_input_data(
    calib_file: str,
    model: torch.nn.Module,
    dataloader: DataLoader,
    get_tensor_func: Optional[Callable] = None,
    inference_func: Optional[Callable] = None,
    model_partition: bool = False,
    context_info: Dict = dict(),
    device: str = "cpu",
) -> None:
    """Create calibration table.

    Examples:
        >>> from juxtapose.mmdeploy.apis.utils import create_calib_input_data
        >>> from juxtapose.mmdeploy.utils import get_calib_filename, load_config
        >>> deploy_cfg = 'configs/mmdet/detection/'
            'detection_tensorrt-int8_dynamic-320x320-1344x1344.py'
        >>> deploy_cfg = load_config(deploy_cfg)[0]
        >>> calib_file = get_calib_filename(deploy_cfg)
        >>> model_cfg = 'mmdetection/configs/fcos/'
            'fcos_r50_caffe_fpn_gn-head_1x_coco.py'
        >>> model_checkpoint = 'checkpoints/'
            'fcos_r50_caffe_fpn_gn-head_1x_coco-821213aa.pth'
        >>> create_calib_input_data(calib_file, deploy_cfg,
            model_cfg, model_checkpoint, device='cuda:0')

    Args:
        calib_file (str): Input calibration file.
        deploy_cfg (str | mmengine.Config): Deployment config.
        model_cfg (str | mmengine.Config): The model config.
        model_checkpoint (str): PyTorch model checkpoint, defaults to `None`.
        dataset_cfg (str | mmengine.Config): Dataset config, defaults to `None`
        dataset_type (str): A string specifying dataset type, e.g.: 'test',
            'val', defaults to 'val'.
        device (str): Specifying the device to run on, defaults to 'cpu'.
    """
    import h5py
    import tqdm

    from juxtapose.mmdeploy.core import RewriterContext, reset_mark_function_count

    backend = "default"

    with h5py.File(calib_file, mode="w") as file:
        calib_data_group = file.create_group("calib_data")

        if not model_partition:
            # create end2end group
            input_data_group = calib_data_group.create_group("end2end")
            input_group = input_data_group.create_group("input")
        for data_id, input_data in enumerate(tqdm.tqdm(dataloader)):
            if not model_partition:
                # save end2end data
                if get_tensor_func is not None:
                    input_tensor = get_tensor_func(input_data)
                else:
                    input_tensor = input_data
                input_ndarray = input_tensor.detach().cpu().numpy()
                input_group.create_dataset(
                    str(data_id),
                    shape=input_ndarray.shape,
                    compression="gzip",
                    compression_opts=4,
                    data=input_ndarray,
                )
            else:
                context_info_ = deepcopy(context_info)
                if "cfg" not in context_info:
                    context_info_["cfg"] = dict()
                context_info_["backend"] = backend
                context_info_["create_calib"] = True
                context_info_["calib_file"] = file
                context_info_["data_id"] = data_id

                with torch.no_grad(), RewriterContext(**context_info_):
                    reset_mark_function_count()
                    if inference_func is not None:
                        inference_func(model, input_data)
                    else:
                        model(input_data)

            file.flush()
